package.preload['projector.projector_proxy_notifies'] = (function (...)
--[[=============================================================================
    Projector Proxy Notification Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--[[
	Notify: INPUT_CHANGED
	Sent to the proxy to indicate the currently selected Input has changed
	Parameters
		bindingID - proxy id of proxy bound to input connection
		input - mod 1000 value of Input Connection ID
--]]
function NOTIFY.INPUT_CHANGED(bindingID, input)
    local tParams = {INPUT = input}
    SendNotify("INPUT_CHANGED", tParams, bindingID)
end
------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
	SendSimpleNotify("ON")
end

function NOTIFY.OFF()
	SendSimpleNotify("OFF")
end
------------------------------ Volume Notification Code ------------------------------
--[[
	Notify: VOLUME_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Volume level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 level uses a percentage scale: 0 - 100
--]]
function NOTIFY.VOLUME_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("VOLUME_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BASS_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Bass level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BASS_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BASS_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: TREBLE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Treble level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.TREBLE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("TREBLE_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BALANCE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Balance level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BALANCE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BALANCE_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: LOUNDENSS_CHANGED
	Sent to the Projector proxy to indicate that the Loudness state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.LOUDNESS_CHANGED(bindingID, output, state)
    local tParams = {LOUDNESS = state, OUTPUT = output + 4000}
    SendNotify("LOUDNESS_CHANGED", tParams, bindingID)		
end

--[[
	Notify: MUTE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Mute state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.MUTE_CHANGED(bindingID, output, state)
    local tParams = {MUTE = state, OUTPUT = output + 4000}
    SendNotify("MUTE_CHANGED", tParams, bindingID)		
end
 end)
package.preload['projector.projector_proxy_commands'] = (function (...)
--[[=============================================================================
    ReceivedFromProxy Code for the Projector Proxy

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.ON(idBinding, tParams)
	gProjectorProxy:prx_ON(tParams)
end

function PRX_CMD.OFF(idBinding, tParams)
	gProjectorProxy:prx_OFF(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gProjectorProxy:prx_SET_INPUT(idBinding, tParams)
end

function PRX_CMD.TV_VIDEO(idBinding, tParams)
	gProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	gProjectorProxy:prx_MUTE_OFF(tParams)
end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	gProjectorProxy:prx_MUTE_ON(tParams)	
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	gProjectorProxy:prx_MUTE_TOGGLE(tParams)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	gProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_UP(tParams)
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_DOWN(tParams)
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_UP(tParams)
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_DOWN(tParams)
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_UP(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.INFO(idBinding, tParams)
	gProjectorProxy:prx_INFO(idBinding, tParams)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	gProjectorProxy:prx_GUIDE(idBinding, tParams)
end

function PRX_CMD.MENU(idBinding, tParams)
	gProjectorProxy:prx_MENU(idBinding, tParams)
end

function PRX_CMD.CANCEL(idBinding, tParams)
	gProjectorProxy:prx_CANCEL(idBinding, tParams)
end

function PRX_CMD.UP(idBinding, tParams)
	gProjectorProxy:prx_UP(idBinding, tParams)
end

function PRX_CMD.DOWN(idBinding, tParams)
	gProjectorProxy:prx_DOWN(idBinding, tParams)
end

function PRX_CMD.LEFT(idBinding, tParams)
	gProjectorProxy:prx_LEFT(idBinding, tParams)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	gProjectorProxy:prx_RIGHT(idBinding, tParams)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_DOWN(idBinding, tParams)
end

function PRX_CMD.START_UP(idBinding, tParams)
	gProjectorProxy:prx_START_UP(idBinding, tParams)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	gProjectorProxy:prx_START_LEFT(idBinding, tParams)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_START_RIGHT(idBinding, tParams)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_UP(idBinding, tParams)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	gProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
end

function PRX_CMD.ENTER(idBinding, tParams)
	gProjectorProxy:prx_ENTER(idBinding, tParams)
end

function PRX_CMD.RECALL(idBinding, tParams)
	gProjectorProxy:prx_RECALL(idBinding, tParams)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
	gProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.NUMBER_0(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_0(idBinding, tParams)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_1(idBinding, tParams)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_2(idBinding, tParams)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_3(idBinding, tParams)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_4(idBinding, tParams)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_5(idBinding, tParams)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_6(idBinding, tParams)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_7(idBinding, tParams)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_8(idBinding, tParams)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_9(idBinding, tParams)
end

function PRX_CMD.STAR(idBinding, tParams)
	gProjectorProxy:prx_STAR(idBinding, tParams)
end

function PRX_CMD.POUND(idBinding, tParams)
	gProjectorProxy:prx_POUND(idBinding, tParams)
end
 end)
package.preload['projector.projector_proxy_class'] = (function (...)
--[[=============================================================================
    Projector Proxy Class Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

DEFAULT_POWER_STATUS_QUERY_INTERVAL = 3         	-- Amount of time to wait BEFORE sending the next power status query, in seconds

ProjectorProxy = inheritsFrom(nil)

function ProjectorProxy:construct(projectorBindingID, bProcessesDeviceMessages, tVolumeRampingTracking, bUsePulseCommandsForVolumeRamping)
	-- member variables
	self._ProjectorBindingID = projectorBindingID
	self._PowerState = ""							--Valid Values: "ON", "OFF", "POWER_ON_SEQUENCE", "POWER_OFF_SEQUENCE"
	self._VolumeIsRamping = false
	self._VolumeRamping = tVolumeRampingTracking	--{["state"] = false, ["mode"] = "",} ||	"state" is boolean, "mode" values: "VOLUME_UP" & "VOLUME_DOWN"
	self._UsePulseCommandsForVolumeRamping = bUsePulseCommandsForVolumeRamping or false
	self._LastVolumeStatusValue = 0	
	self._MenuNavigationInProgress = false
     self._MenuNavigationMode = ""
	self._MenuNavigationProxyID = ""	
	self._CurrentlySelectedInput = ""
	self._ProcessesDeviceMessages = bProcessesDeviceMessages
	self._ControlMethod = ""						--Valid Values: "NETWORK", "SERIAL", "IR" 
	self._PowerStatusQueryDelayS = DEFAULT_COMMAND_RESPONSE_INTERVAL
end

function ProjectorProxy:Initialize(PowerStatusQueryDelayInterval, CallbackParam)
	if (PowerStatusQueryDelayInterval ~= nil) then
		self._PowerStatusQueryDelayS = PowerStatusQueryDelayInterval
	end
	self._PowerStatusQueryDelayTimer = c4_timer:new("PowerStatusQueryDelay", self._PowerStatusQueryDelayS, "SECONDS", ProjectorProxy.OnPowerStatusQueryDelayTimerExpired, false, self)
end

-- Note the '.' instead of the ':'
function ProjectorProxy.OnPowerStatusQueryDelayTimerExpired(Instance)
	LogTrace("Power Status Query Timer expired")
	Instance._PowerStatusQueryDelayTimer:KillTimer()
	
	if (GetDevicePowerStatus ~= nil and type(GetDevicePowerStatus) == "function") then
		GetDevicePowerStatus(true)  --"true" sends command as a priority command
	end
end

------------------------------------------------------------------------
-- Projector Proxy Commands (PRX_CMD)
------------------------------------------------------------------------
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_ON(tParams)
--[[
	if (self._PowerState ~= nil) then
		if (self._PowerState == 'ON') or (self._PowerState == 'POWER_ON_SEQUENCE') then 
			LogInfo("Power State is '" .. self._PowerState .. "', exiting ON")
			return 
		end	
	end	
	if (self._ProcessesDeviceMessages == false) then
		self._PowerState = 'ON'
	else	
		self._PowerState = 'POWER_ON_SEQUENCE'
	end	
--]]
    if (gControlMethod == "IR") then			
		NOTIFY.ON()			
    end	
    
    ON()
end

function ProjectorProxy:prx_OFF(tParams)
    if (self._ProcessesDeviceMessages == false) then
	   self._PowerState = 'OFF'
    else	
	   self._PowerState = 'POWER_OFF_SEQUENCE'
    end	
    if (gControlMethod == "IR") then			
	   NOTIFY.OFF()			
    end 	
    OFF()
  
    self._CurrentlySelectedInput = -1
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_SET_INPUT(idBinding, tParams)
	local input = tonumber(tParams["INPUT"] % 1000)
	self._CurrentlySelectedInput = input
	if (gControlMethod == "IR") then			
		NOTIFY.INPUT_CHANGED(self._ProjectorBindingID, input)				
	end 		

	SET_INPUT(idBinding, input)
end

function ProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "TV_VIDEO")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_MUTE_OFF(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "False")
	end 		
	MUTE_OFF()
end

function ProjectorProxy:prx_MUTE_ON(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "True")
	end 		
	MUTE_ON()
end

function ProjectorProxy:prx_MUTE_TOGGLE(tParams)
	MUTE_TOGGLE()
end

function ProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
	local c4VolumeLevel = tonumber(tParams['LEVEL'])
	SET_VOLUME_LEVEL(c4VolumeLevel)
end

function ProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
	PULSE_VOL_DOWN()
end

function ProjectorProxy:prx_PULSE_VOL_UP(tParams)
	PULSE_VOL_UP()
end

function ProjectorProxy:prx_START_VOL_DOWN(tParams)
	self:ChangeVolume("START_VOL_DOWN")
end

function ProjectorProxy:prx_START_VOL_UP(tParams)
	self:ChangeVolume("START_VOL_UP")
end

function ProjectorProxy:prx_STOP_VOL_DOWN(tParams)
	self:ChangeVolume("STOP_VOL_DOWN")
end

function ProjectorProxy:prx_STOP_VOL_UP(tParams)
	self:ChangeVolume("STOP_VOL_UP")
end

---------------------- Volume Helper Functions ----------------------
function ProjectorProxy:ChangeVolume(command_name)
	if (command_name == "STOP_VOL_UP") or (command_name == "STOP_VOL_DOWN") then
		self._VolumeIsRamping = false
		self._VolumeRamping.state = false
		self._VolumeRamping.mode = ""
	elseif (command_name == "START_VOL_UP") then 
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_UP" 
		PULSE_VOL_UP()	
	elseif (command_name == "START_VOL_DOWN") then 	
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_DOWN"	
		PULSE_VOL_DOWN()		
	else
		LogWarn(command_name .. " not handled in ChangeVolume()")
	end
end

function ProjectorProxy:ContinueVolumeRamping()
	local command
	if (gControlMethod == "IR") then   
		if (self._VolumeRamping.mode == "VOLUME_UP") then
			PULSE_VOL_UP()	
		elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
			PULSE_VOL_DOWN()	
		else
			LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		end	
	else
	     if (self._UsePulseCommandsForVolumeRamping) then
		    if (self._VolumeRamping.mode == "VOLUME_UP") then
			 PULSE_VOL_UP()	
		    elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
			 PULSE_VOL_DOWN()	
		    else
			 LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		    end	
		else
		    local volume = self._LastVolumeStatusValue
		    local deviceVolumeLevel = self:GetNextVolumeCurveValue(volume)
		    if (deviceVolumeLevel ~= nil) then
			    self._LastVolumeStatusValue = deviceVolumeLevel
			    SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)                                 
		    else
			    LogWarn("ContinueVolumeRamping() next value is nil")
			    return
		    end
	     end
	end 
end

function ProjectorProxy:GetNextVolumeCurveValue(volume)
	local i, point
	volume=tonumber(volume)
	if (self._VolumeRamping.mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		LogWarn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ProjectorProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    local curveV = {}
    curveV.__index = curveV

    function curveV:new(min, max, base)
	 local len = max-min
	 local logBase = math.log(base)
	 local baseM1 = base-1

	 local instance = {
	   min = min,
	   max = max,
	   len = len,
	   base = base,
	   logBase = logBase,
	   baseM1 = baseM1,
	   toNormal = function(x)
		return (x-min)/len
	   end,
	   fromNormal = function(x)
		return min+x*len
	   end,
	   value = function(x)
		return math.log(x*baseM1+1)/logBase
	   end,
	   invert = function(x)
		return (math.exp(x*logBase)-1)/baseM1
	   end
	 }
	 return setmetatable(instance, self)
    end

    function curveV:list(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))

	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	 end
    end
    
    function curveV:create_curve(from, to, steps)
	 local fromI = self.invert(self.toNormal(from))
	 local toI = self.invert(self.toNormal(to))
	 
      local tCurve = {}
	 for i = 1, steps do
	   --print(i, self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	   --print(i, math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1)))))
	    
	    local x = math.floor(self.fromNormal(self.value(fromI+(i-1)*(toI-fromI)/(steps-1))))
	    
	    if (has_value(tCurve, x) == false) then
		  table.insert(tCurve, x)
		  print(i, x)
	    end
	 end
	 
	 return tCurve
    end
    
    

    -- min, max, base of log (must be > 1), try some for the best results
    local a = curveV:new(minDeviceLevel, maxDeviceLevel, slope)

    -- from, to, steps
    --a:list(minDeviceLevel, maxDeviceLevel, steps)
    
    local t = a:create_curve(minDeviceLevel, maxDeviceLevel, steps)
    
    return t

end

function has_value(tab, val)
    for index, value in ipairs(tab) do
        if value == val then
            return true
        end
    end

    return false
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	  local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --LogInfo("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --LogInfo("vl = " .. vl)
	  vl= tonumber_loc(("%.".."0".."f"):format(vl))
	  --LogInfo("vl new = " .. vl)
	  LogInfo("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_PULSE_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_START_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_START_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_PRESET_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_UP")
end

function ProjectorProxy:prx_PRESET_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_DOWN")
end

function ProjectorProxy:prx_SET_CHANNEL(idBinding, tParams)
    local channel= tParams['CHANNEL']
	SET_CHANNEL(idBinding, channel)
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_INFO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "INFO")
end

function ProjectorProxy:prx_GUIDE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "GUIDE")
end

function ProjectorProxy:prx_MENU(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "MENU")
end

function ProjectorProxy:prx_CANCEL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "CANCEL")
end

function ProjectorProxy:prx_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "UP")
end

function ProjectorProxy:prx_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "DOWN")
end

function ProjectorProxy:prx_LEFT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "LEFT")
end

function ProjectorProxy:prx_RIGHT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RIGHT")
end

function ProjectorProxy:prx_START_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_DOWN")
end

function ProjectorProxy:prx_START_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_UP")
end

function ProjectorProxy:prx_START_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_LEFT")
end

function ProjectorProxy:prx_START_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_RIGHT")
end

function ProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_DOWN")
end

function ProjectorProxy:prx_STOP_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_UP")
end

function ProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_LEFT")
end

function ProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_RIGHT")
end

function ProjectorProxy:prx_ENTER(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "ENTER")
end

function ProjectorProxy:prx_RECALL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RECALL")
end

function ProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "OPEN_CLOSE")
end

function ProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_A")
end

function ProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_B")
end

function ProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_C")
end

function ProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_D")
end

---------------------- Menu Navigation Helper Functions ----------------------
function ProjectorProxy:NavigateMenu(idBinding, command_name)
	if (command_name == "STOP_UP") 
				or (command_name == "STOP_DOWN") 
				or (command_name == "STOP_LEFT") 
				or (command_name == "STOP_RIGHT") then
		self._MenuNavigationInProgress = false
		self._MenuNavigationMode = ""
		self._MenuNavigationProxyID = ""
		return
	elseif (command_name == "START_UP") then 
		self._MenuNavigationMode = "UP"
	elseif (command_name == "START_DOWN") then 	
		self._MenuNavigationMode = "DOWN"	
	elseif (command_name == "START_LEFT") then 
		self._MenuNavigationMode = "LEFT"	
     elseif (command_name == "START_RIGHT") then 
		self._MenuNavigationMode = "RIGHT"
	else
		LogWarn(command_name .. " not handled in NavigateMenu()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, self._MenuNavigationMode)
	self._MenuNavigationInProgress = true
	self._MenuNavigationProxyID = idBinding
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_NUMBER_0(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_0")
end

function ProjectorProxy:prx_NUMBER_1(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_1")
end

function ProjectorProxy:prx_NUMBER_2(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_2")
end

function ProjectorProxy:prx_NUMBER_3(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_3")
end

function ProjectorProxy:prx_NUMBER_4(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_4")
end

function ProjectorProxy:prx_NUMBER_5(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_5")
end

function ProjectorProxy:prx_NUMBER_6(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_6")
end

function ProjectorProxy:prx_NUMBER_7(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_7")
end

function ProjectorProxy:prx_NUMBER_8(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_8")
end

function ProjectorProxy:prx_NUMBER_9(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_9")
end

function ProjectorProxy:prx_STAR(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "STAR")
end

function ProjectorProxy:prx_POUND(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "POUND")
end

------------------------------------------------------------------------
-- Projector Proxy Notifies
------------------------------------------------------------------------

function ProjectorProxy:dev_InputChanged(bindingID, input)
	NOTIFY.INPUT_CHANGED(bindingID, input)
end

function ProjectorProxy:dev_PowerOn()
	self._PowerState = "ON"
	NOTIFY.ON()	
end

function ProjectorProxy:dev_PowerOff()
	self._PowerState = "OFF"
	NOTIFY.OFF()
end

function ProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	NOTIFY.VOLUME_LEVEL_CHANGED(self._ProjectorBindingID, output, c4Level)	
	
	if (self._VolumeIsRamping) then
		--do nothing
		--during volume ramping, LastVolumeStatusValue is set in ContinueVolumeRamping()
	else
		self._LastVolumeStatusValue = deviceLevel
	end	
end

function ProjectorProxy:dev_MuteChanged(output, state)
	NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, output, state)
end		

function ProjectorProxy:dev_BassLevelChanged(output, level)
	NOTIFY.BASS_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_TrebleLevelChanged(output, level)
	NOTIFY.TREBLE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_BalanceLevelChanged(output, level)
	NOTIFY.BALANCE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_LoudnessChanged(output, state)
	NOTIFY.LOUDNESS_CHANGED(self._ProjectorBindingID, output, state)
end

function ProjectorProxy:dev_ChannelStatus(channel)
	NOTIFY.CHANNEL_CHANGED(self._ProjectorBindingID, channel)
end end)
package.preload['common.c4_url_connection'] = (function (...)
--[[=============================================================================
    Base for a url connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_url_connection = "2016.01.08"
end

UrlConnectionBase = inheritsFrom(DeviceConnectionBase)

function UrlConnectionBase:construct(Url)
	self.superClass():construct()
	self._Url = Url
end

function UrlConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "URL"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	OnURLConnectionChanged()
end

function UrlConnectionBase:ControlMethod()
	return "URL"
end

function UrlConnectionBase:SetUrl(Url)
	self._Url = Url
end

function UrlConnectionBase:SendCommand(sCommand, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(self._Url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(self._Url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:SendCommandUrl(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlPost(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlGet(url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlGet(url, sHeader)
		else
			ticketId = C4:urlGet(url)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:ReceivedAsync(ticketId, sData, responseCode, tHeaders)
	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(sData))
	LogTrace(tHeaders)
	local tMessage = {
		["ticketId"] = ticketId,
		["sData"] = sData,
		["responseCode"] = responseCode,
		["tHeaders"] = tHeaders
	}
	
	status, err = pcall(HandleMessage, tMessage)
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

function ConnectURL()
	gIsUrlConnected = true
	SetControlMethod()
end

function DisconnectURL()
	gIsUrlConnected = false
	SetControlMethod()
end
 end)
package.preload['connections'] = (function (...)
--[[=============================================================================
    Functions for managing the status of the drivers bindings and connection state
 
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_url_connection"

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
    TEMPLATE_VERSION.connections = "2015.03.31"
end

-- constants
COM_USE_ACK = true
COM_COMMAND_DELAY_MILLISECONDS = 250
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4

NETWORK_PORT = 53595
IR_BINDING_ID = 2
SERIAL_BINDING_ID = 1
NETWORK_BINDING_ID = 6001

function ON_DRIVER_LATEINIT.connections()
    local xml = C4:GetDriverConfigInfo("connections")
    for c in string.gmatch(xml, '<connection>(.-)</connection>') do
	   local id, port = string.match(c, '<id>(%d+)</id>.-<classname>TCP</classname>.-<number>(%d+)</number>')
	   if id ~= nil and port ~= nil then
		  NETWORK_BINDING_ID = tonumber(id)
		  NETWORK_PORT	= tonumber(port)
		  print(NETWORK_BINDING_ID, NETWORK_PORT)
	   end
    end
  
    GetMACAddress()
    GetIPAddress()

end

function GetMACAddress()
    if (LuaC4Object.GetDeviceMAC) then
	   local mac = C4:GetDeviceMAC(NETWORK_BINDING_ID) or ""
	   
	   if (mac ~= "") then
		  UpdateProperty("MAC Address", mac)
	   end
	   
	   return mac
    end
    
    return ""
end

function GetIPAddress()
    if (LuaC4Object.GetBindingAddress) then
	   local addr = C4:GetBindingAddress(NETWORK_BINDING_ID) or ""
	   
	   if (addr ~= "") then
		  UpdateProperty("IP Address", addr)
	   end
	   
	   return addr
    end
    
    return ""
end


--[[=============================================================================
    OnSerialConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when a serial binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnSerialConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnIRConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when an IR binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnIRConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnNetworkConnectionChanged(idBinding, bIsBound)
  
    Description:
    Function called when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkConnectionChanged(idBinding, bIsBound)
    if (idBinding == NETWORK_BINDING_ID) and bIsBound then
	   gNetworkInitTimer:StartTimer()
    end
end

function OnNetworkInitTimerExpired()
    gNetworkInitTimer:KillTimer()
    GetMACAddress()
    GetIPAddress()
end

--[[=============================================================================
    OnNetworkStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Called when the network connection status changes. Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnNetworkStatusChanged(idBinding, nPort, sStatus)
    if (idBinding == NETWORK_BINDING_ID) then
	   if (sStatus	== "ONLINE") then
		  if (Properties["MAC Address"] == nil or Properties["MAC Address"] == "") then
			 gNetworkInitTimer:StartTimer()
		  end
		  
		  --if (gProjectorProxy._PowerState == "OFF") then gQueryPowerDelayTimer:StartTimer(2, "SECONDS") end
		  
	   else
		  --TV does not respond when OFF
		  --if timer is not killed here, then the driver will continually try to reconnect when TV is OFF
		  gCon._NetworkReconnectTimer:KillTimer()	
		  
		  --TV may have been turned off by factory remote or due to lack of signal...
		  if (gProjectorProxy._PowerState ~= "OFF") then
			 gProjectorProxy:dev_PowerOff()
		  end
	   end				 
    else
    end	
end

--[[=============================================================================
    OnURLConnectionChanged(url)
  
    Description:
    Function called when the c4_url_connection is created.
  
    Parameters:
    url - url used by the url connection.
  
    Returns:
    None
===============================================================================]]
function OnURLConnectionChanged(url)
	
end

--[[=============================================================================
    DoEvents()
  
    Description:
    Called from OnSendTimeExpired in the DeviceConnectionBase.
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function DoEvents()
	if (gProjectorProxy._VolumeIsRamping == true) then
		if (gProjectorProxy._VolumeRamping.state == true) then gProjectorProxy:ContinueVolumeRamping() end
	end
end

--[[=============================================================================
    SendKeepAlivePollingCommand()
  
    Description:
    Sends a driver specific polling command to the connected system
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function SendKeepAlivePollingCommand()
    LogTrace("SendKeepAlivePollingCommand()")
	
    --local command = "version ?" 
    local command = "timer ?" 
    LogTrace("command = " .. command)
    PackAndQueueCommand("SendKeepAlivePollingCommand", command)		
end
 end)
package.preload['proxy_commands'] = (function (...)
--[[=============================================================================
    Commands received from the Projector proxy (ReceivedFromProxy)

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: ON
	Parameters:
--]]

function ON()
    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
          
    if (gProjectorProxy._PowerState ~= nil) then
	   if (gProjectorProxy._PowerState == 'ON')  then 
		  if ((gControlMethod == "IR") or (gProjectorProxy._ProcessesDeviceMessages == false)) then
			 command_delay = tonumber(Properties["Power On Delay Seconds"])
			 delay_units = "SECONDS"
		  else
			 command_delay = tonumber(Properties["Command Delay Milliseconds"])
			 delay_units = "MILLISECONDS"
		  end
	   elseif (gProjectorProxy._PowerState == 'POWER_ON_SEQUENCE') then
		  command_delay = tonumber(Properties["Power On Delay Seconds"])
		  delay_units = "SECONDS"
	   end	
    else
	   command_delay = tonumber(Properties["Power On Delay Seconds"])
	   delay_units = "SECONDS"
    end	

    gPowerOnTimer:StartTimer(90, delay_units)
    
    if gPowerOffTimer:TimerStarted() then
	   --quick power on AFTER power off, need TV to turn off before sending on. This occurs in OnPowerOffTimer...
	   return
    end
    
    if (gControlMethod == nil) then sendWOL() end	--AFTER DRIVER UPDATED gControlMethod may not be set...

    local command
    if (gControlMethod == "IR") then
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["ON"]
		--LogTrace("command = " .. command)
		PackAndQueueCommand("ON", command, command_delay, delay_units)
		
		return
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["ON"]
	   PackAndQueueCommand("ON", command, command_delay, delay_units) --sending twice, depends on if the network managment is set to OFF
    elseif (gControlMethod == "Network") then
	   gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)
	   if (Properties["Power ON Method"] == "Wake On LAN") then
		  sendWOL()
		  return
	   else
		  command = CMDS["ON"]
	   end
    else
	   LogTrace("Exiting ON(), gControlMethod not set.")
	   return
    end
    
    --LogTrace("command = " .. command)
    PackAndQueueCommand("ON", command, command_delay, delay_units)
    --GetLampHours()
    --GetDevicePowerStatus()
    
end

function sendWOL()
    if (LuaC4Object.SendWOL) then	--C4:SendWOL() added in OS 2.10
	   local mac_string = format_mac_string()
	   
	   for i = 1,5 do	   
		  C4:SendWOL(mac_string, 9)
	   end
    else
	   local mac_hex = format_mac_hex()
	   C4:CreateNetworkConnection (6002, "255.255.255.255", 9) --create network connection 6002 for Wake on Lan
	   C4:NetConnect(6002, 9, "UDP") --connect to connection 6002 and make it UDP for Wake On Lan
	   
	   local packet = string.rep(string.char(255), 6) .. string.rep(mac_hex, 16) -- Build 'magic packet'. 
	   --hexdump (packet) 
	   for i = 1,5 do	  
		  C4:SendToNetwork (6002, 9, packet)
	   end
    end
    
end

function format_mac_string()
    local s = ""
	s = Properties["MAC Address"] or ""
	if (s == "") then s = GetMACAddress() end

	local mac = s:gsub("%X", "")
	return mac
end

function format_mac_hex()
    local s = ""
	s = Properties["MAC Address"] or ""
	if (s == "") then s = GetMACAddress() end
	
	local mac = s:gsub("%X", " ")
	return tohex(mac)
end

--[[
	Proxy Command: OFF
	Parameters:
--]]

gPowerOffStartTime = 0

function OFF()
    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    
    --in case of a quick power on / power off sequence kill the power on timers in progress...
    gPowerOnTimer:KillTimer()
    gPowerOffTimer:StartTimer(tonumber(Properties["Power Off Delay Seconds"]), "SECONDS")
    gPowerOffStartTime = os.time()
    
    local command 
    if (gControlMethod == "IR") then
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["OFF"]
		--LogTrace("command = " .. command)
		PackAndQueueCommand("OFF", command, command_delay, delay_units)
		
		return
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["OFF"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["OFF"]
    else
	   LogTrace("Exiting OFF(), gControlMethod not set.")
	   return
    end
    
    --LogTrace("command = " .. command)
    PackAndQueueCommand("OFF", command, command_delay, delay_units)
    --GetDevicePowerStatus() 

end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: SET_INPUT
	Parameters:
		idBinding: proxybindingid of proxy bound to input connection
		output: mod 1000 value of Output Connection id	
		input: mod 1000 value of Input Connection id
--]]
function SET_INPUT(idBinding, input)
    local command
    if (gControlMethod == "IR") then		
	   command = tInputCommandMap_IR[tInputConnMapByID[input].Name]	
    elseif (gControlMethod == "Serial") then
	   command = tInputCommandMap_Serial[tInputConnMapByID[input].Name] 
    elseif (gControlMethod == "Network") then
	   command = tInputCommandMap[tInputConnMapByID[input].Name] 
    else
	   LogTrace("Exiting SET_INPUT(), gControlMethod not set.")
	   return
    end 		

    local delay, delay_units
    if (gPowerOffTimer:TimerStarted()) then
	   delay = 120	--tonumber(Properties["Power On Delay Seconds"])
	   delay_units = "SECONDS"
	   --elseif (gPowerOnTimer:TimerStarted()) then
    elseif (gProjector_Operating_Status ~= "POWER ON") then
	   delay = 120	--tonumber(Properties["Power On Delay Seconds"])
	   delay_units = "SECONDS"
    else
	   --delay = tonumber(Properties["Command Delay Milliseconds"])
	   --delay_units = "MILLISECONDS"
	   delay = 1	
	   delay_units = "SECONDS"
    end
    
    gSetInputDelayTimer:StartTimer(delay, delay_units)
    gSetInputDelayTimer._CallbackParam = command	
	
end

function OnSetInputDelayTimerExpired(command)
    LogTrace("OnSetInputDelayTimerExpired, command = " .. command)
    gSetInputDelayTimer:KillTimer()
    
    PackAndQueueCommand("SET_INPUT", command)
    
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: MUTE_OFF
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_OFF()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["MUTE_OFF"]	
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_OFF"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["MUTE_OFF"]
    else
	   LogTrace("Exiting MUTE_OFF(), gControlMethod not set.")
	   return
    end 	
	
    --LogTrace("command = " .. command)
    PackAndQueueCommand("MUTE_OFF", command)
end

--[[
	Proxy Command: MUTE_ON
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_ON()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_ON"]	
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_ON"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["MUTE_ON"]
    else
	   LogTrace("Exiting MUTE_ON(), gControlMethod not set.")
	   return
    end 	
	
    --LogTrace("command = " .. command)
    PackAndQueueCommand("MUTE_ON", command)		
end

--[[
	Proxy Command: MUTE_TOGGLE
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_TOGGLE()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_TOGGLE"]	
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["MUTE_TOGGLE"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["MUTE_TOGLE"]
    else
	   LogTrace("Exiting MUTE_TOGGLE(), gControlMethod not set.")
	   return
	end 
	
	--LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_TOGGLE", command)
end

--[[
	Proxy Command: SET_VOLUME_LEVEL
	Parameters:
		output: mod 1000 value of Output Connection id	
		c4VolumeLevel: volume level to be set represented in Control4 scale (0-100)
--]]
function SET_VOLUME_LEVEL(c4VolumeLevel)
	local minDeviceLevel = MIN_DEVICE_LEVEL
	local maxDeviceLevel = MAX_DEVICE_LEVEL
	local deviceVolumeLevel = ConvertVolumeToDevice(c4VolumeLevel, minDeviceLevel, maxDeviceLevel)	
	
	---- TODO: uncomment and edit string padding, if required, based upon the protocol specification
    --deviceVolumeLevel = string.rep("0", 2 - string.len(deviceVolumeLevel)) .. deviceVolumeLevel
	
	LogInfo('deviceVolumeLevel: ' .. deviceVolumeLevel)
  
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	--LogTrace("command = " .. command)
	PackAndQueueCommand("SET_VOLUME_LEVEL", command)
end

--[[
	Helper Function: SET_VOLUME_LEVEL_DEVICE
	Parameters:
		output: mod 1000 value of Output Connection id	
		deviceVolumeLevel: volume level to be set represented in device scale (as sepcified in the device's control protocol)
--]]
function SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)
	--Called from ContinueVolumeRamping()
	
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	
	--LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("SET_VOLUME_LEVEL_DEVICE", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_DOWN
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_DOWN()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_DOWN"]	
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["PULSE_VOL_DOWN"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["PULSE_VOL_DOWN"]
    else
	   LogTrace("Exiting PULSE_VOL_DOWN(), gControlMethod not set.")
	   return
    end 	
	
    --LogTrace("command = " .. command)
    local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
    PackAndQueueCommand("PULSE_VOL_DOWN", command, command_delay)
end

--[[
	Proxy Command: PULSE_VOL_UP
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_UP()
	local command
	if (gControlMethod == "IR") then	
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_UP"]	
    elseif (gControlMethod == "Serial") then
	   command = CMDS_Serial["PULSE_VOL_UP"]
    elseif (gControlMethod == "Network") then
	   command = CMDS["PULSE_VOL_UP"]
    else
	   LogTrace("Exiting PULSE_VOL_UP(), gControlMethod not set.")
	   return
    end 	
    		
    --LogTrace("command = " .. command)
    local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
    PackAndQueueCommand("PULSE_VOL_UP", command, command_delay)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	Helper Function: SEND_COMMAND_FROM_COMMAND_TABLE
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be sent
--]]
function SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, command_name)
    LogTrace("SEND_COMMAND_FROM_COMMAND_TABLE(), idBinding=" .. idBinding .. ", command_name=" .. command_name)
	
	-- TODO: create packet/command to send to the device
	local command = GetCommandFromCommandTable(idBinding, command_name)
	
	if (command == nil) then
		LogTrace("command is nil")
	else
		LogTrace("command = " .. command)
		PackAndQueueCommand(command_name, command)
	end			
end

--[[
	Helper Function: GetCommandFromCommandTable
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be returned
--]]
function GetCommandFromCommandTable(idBinding, command_name)
	LogTrace("GetCommand()")
	local t = {}
	
	-- TODO: create appropriate commands table structure
	
	if (gControlMethod == "IR") then
		t = CMDS_IR
     elseif (gControlMethod == "Serial") then
	     t = CMDS_Serial
	else
		t = CMDS
	end	

	if (t[idBinding] ~= nil) then
	   if (t[idBinding][command_name] ~= nil) then
		  return t[idBinding][command_name]
	   end
	end
	
	if (t[command_name] ~= nil) then
		return t[command_name]
	else
		LogWarn('GetCommandFromCommandTable: command not defined - '.. command_name)
		return nil
	end	
	
end

--[[
	Helper Function: GetDeviceVolumeStatus
--]]
function GetDeviceVolumeStatus()
    LogTrace("GetDeviceVolumeStatus()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_VOLUME_STATUS"]
	
	local command
	if (gControlMethod == "Serial") then
	   command = CMDS_Serial["GET_VOLUME_STATUS"]
     else
	   command = CMDS["GET_VOLUME_STATUS"]
	end
	
	--LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Volume", command)
	
	if (gControlMethod == "Serial") then
	   command = CMDS_Serial["GET_MUTE_STATUS"]
     else
	   command = CMDS["GET_MUTE_STATUS"]
	end
	
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Mute", command)	
end

--[[
	Helper Function: GetLampHours
--]]
function GetLampHours()
    LogTrace("GetLampHours()")
	
	local command
	if (gControlMethod == "Serial") then
	   command = CMDS_Serial["GET_LAMP_HOURS"]
     else
	   command = CMDS["GET_LAMP_HOURS"]
	end
    
	--LogTrace("command = " .. command)
	PackAndQueueCommand("GetLampHours", command)
end

--[[
	Helper Function: GetDevicePowerStatus
--]]
function GetDevicePowerStatus(...)
	local bPriority = select(1, ...) or false
     LogTrace("GetDevicePowerStatus()")
	
	local command
	if (gControlMethod == "Serial") then
	   command = CMDS_Serial["GET_POWER_STATUS"]
     else
	   command = CMDS["GET_POWER_STATUS"]
	end
	--LogTrace("command = " .. command)
	
	if (bPriority) then
		PackAndQueuePriority1Command("GetDevicePowerStatus", command)
	else
		PackAndQueueCommand("GetDevicePowerStatus", command)
	end
end

function GetDeviceInputStatus()
     LogTrace("GetDeviceVolumeStatus()")

	local command
	if (gControlMethod == "Serial") then
	   command = CMDS_Serial["GET_INPUT_STATUS"]
     else
	   command = CMDS["GET_INPUT_STATUS"]
	end
	
	--LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceInputStatus", command)
	
end
	
 end)
package.preload['properties'] = (function (...)
--[[=============================================================================
    Properties Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

function ON_PROPERTY_CHANGED.NetworkKeepAliveIntervalSeconds(propertyValue)
	gNetworkKeepAliveInterval = tonumber(propertyValue)

end

function ON_PROPERTY_CHANGED.RampingMethod(propertyValue)
    if (gInitializingDriver == true) then return end
    
    if (propertyValue == "Pulse Commands") then 
	   gProjectorProxy._UsePulseCommandsForVolumeRamping = true 
    else
	   gProjectorProxy._UsePulseCommandsForVolumeRamping = false
    end
    
end

function ON_PROPERTY_CHANGED.VolumeRampSteps(propertyValue)
    if (gInitializingDriver == true) then return end
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    
end

function ON_PROPERTY_CHANGED.VolumeRampSlope(propertyValue)
    if (gInitializingDriver == true) then return end
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)

	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end

 end)
package.preload['projector_init'] = (function (...)
--[[=============================================================================
    PROJECTOR Protocol Initialization Functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end


PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_EARLY_INIT.projector_init()

end

function ON_DRIVER_INIT.projector_init()
	--LogTrace("ON_DRIVER_INIT.ProtocolDeclarations()")
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
end

function ON_DRIVER_LATEINIT.projector_init()
end

function PROTOCOL_DECLARATIONS.SetGlobals()
	LogTrace("PROTOCOL_DECLARATIONS.SetGlobals()")

	--[[
	TODO: Determine projector operating conditions to be exposed for Composer Programming.
		Most projectors will report status, i.e. Warm Up,Cool Down and Error, etc.
		Lamp usage may also be reported. The sample variables below are referenced in the
		DEV_MSG.POWER() function in the device_messages.lua file.
	--]]
	C4:AddVariable('Projector_Operating_Status',0,'STRING',false,false)	
	C4:AddVariable('Lamp_Hours',0,'STRING',false,false)
	gProjector_Operating_Status = ""
	gLamp_Hours = ""
end
	
function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--index:  Proxy Command Name
		--value:  IR Code ID	 
		--Power
		["ON"]             = "",
		["OFF"]            = "",
		
		--Input Toggle
		["INPUT_TOGGLE"]="",
		["TV_VIDEO"] 		= "",
		
		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",	
		["RECALL"]         = "",
		["PREV"]           = "",
		
		--Digits
		["NUMBER_0"]       = "", 	
		["NUMBER_1"]       = "",		
		["NUMBER_2"]       = "",	
		["NUMBER_3"]       = "",	
		["NUMBER_4"]       = "",	
		["NUMBER_5"]       = "",	
		["NUMBER_6"]       = "",	
		["NUMBER_7"]       = "",	
		["NUMBER_8"]       = "",	
		["NUMBER_9"]       = "",
		["STAR"]           = "",
		["DOT"]            = "",
		
		--Volume
		["MUTE_ON"]        = "",
		["MUTE_OFF"]       = "",
		["MUTE_TOGGLE"]    = "",		
		["PULSE_VOL_DOWN"] = "",
		["PULSE_VOL_UP"]   = "",		
		["SET_VOLUME_LEVEL"] = "",		
		
	}			
	
	CMDS_IR[PROJECTOR_PROXY_BINDINGID] = {}	
	
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_ADCP()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_ADCP()")

	CMDS = {
		--index:  Proxy Command Name
		--value:  Protocol Command Data
		
		--Power
		["ON"]             = 'power "on"',
		["OFF"]            = 'power "off"',
		["POWER_TOGGLE"]   = 'key "power"',
		
		--Input Toggle
		["INPUT_TOGGLE"]='',
		["TV_VIDEO"]		= '',
		
		--Menu
		["INFO"] = '',	--Display
		["GUIDE"] = '',
		["MENU"] = 'key "menu"',
		["CANCEL"] = 'key "menu"',
		["UP"] = 'key "up"',
		["DOWN"] = 'key "down"',
		["LEFT"] = 'key "left"',
		["RIGHT"] = 'key "right"',
		["ENTER"] = 'key "enter"',	
		["RECALL"]         = '',
		["PREV"]           = '',
		
		--Digits
		["NUMBER_0"]       = '', 	
		["NUMBER_1"]       = '',		
		["NUMBER_2"]       = '',	
		["NUMBER_3"]       = '',	
		["NUMBER_4"]       = '',	
		["NUMBER_5"]       = '',	
		["NUMBER_6"]       = '',	
		["NUMBER_7"]       = '',	
		["NUMBER_8"]       = '',	
		["NUMBER_9"]       = '',
		["STAR"]           = '',
		["DOT"]            = '',
		
		--Volume
		["MUTE_ON"]        = '',
		["MUTE_OFF"]       = '',
		["MUTE_TOGGLE"]    = 'key "muting"',	
		["PULSE_VOL_DOWN"] = 'key "vol-"',
		["PULSE_VOL_UP"]   = 'key "vol+"',		
		["SET_VOLUME_LEVEL"] = '',		
	
		--Status Query
		["GET_VOLUME_STATUS"]	= '',
		["GET_MUTE_STATUS"]		= '',
		["GET_POWER_STATUS"]	= 'power_status ?',
		["GET_LAMP_HOURS"]		= 'timer ?',
		['GET_INPUT_STATUS']	= 'input ?',
		
	}
	
	CMDS[PROJECTOR_PROXY_BINDINGID] = {}
	
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()")

	CMDS_Serial = {
	   --index:  Proxy Command Name
	   --value:  Protocol Command Data

	   --Power
	   ["ON"]             = tohex('01 30 00 00 01'),
	   ["OFF"]            = tohex('01 30 00 00 00'),

	   --Input Toggle
	   ["INPUT_TOGGLE"]="",
	   ["TV_VIDEO"]		= "",

	   --Menu
	   ['MENU']=tohex('17 29 00 00 00'),
	   ['UP']=tohex('17 35 00 00 00'),
	   ['DOWN']=tohex('17 36 00 00 00'),
	   ['LEFT']=tohex('17 34 00 00 00'),
	   ['RIGHT']=tohex('17 33 00 00 00'),
	   ['ENTER']=tohex('17 5A 00 00 00'),
	   ["INFO"] = "",	--Display
	   ["GUIDE"] = "",
	   ["CANCEL"] = "",
	   ["RECALL"]         = "",
	   ["PREV"]           = "",

	   --Digits
	   ["NUMBER_0"]       = "", 	
	   ["NUMBER_1"]       = "",		
	   ["NUMBER_2"]       = "",	
	   ["NUMBER_3"]       = "",	
	   ["NUMBER_4"]       = "",	
	   ["NUMBER_5"]       = "",	
	   ["NUMBER_6"]       = "",	
	   ["NUMBER_7"]       = "",	
	   ["NUMBER_8"]       = "",	
	   ["NUMBER_9"]       = "",
	   ["STAR"]           = "",
	   ["DOT"]            = "",

	   --Volume
	   ["MUTE_ON"]        = "",
	   ["MUTE_OFF"]       = "",
	   ["MUTE_TOGGLE"]    = "",	
	   ["PULSE_VOL_DOWN"] = "",
	   ["PULSE_VOL_UP"]   = "",		
	   ["SET_VOLUME_LEVEL"] = "",		

	   --Status Query
	   ["GET_VOLUME_STATUS"]	= "",
	   ["GET_MUTE_STATUS"]		= "",
	   ["GET_POWER_STATUS"]	= tohex('01 02 01 00 00'),
	   ["GET_LAMP_HOURS"]		= tohex('01 13 01 00 00'),
	   ['GET_PROJECTOR_ERROR_STATUS']= tohex('01 01 01 00 00'),
	   ['GET_INPUT_STATUS']= tohex('00 01 01 00 00'),
	}
	
	CMDS_Serial[PROJECTOR_PROXY_BINDINGID] = {}	
	
	
end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	LogTrace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	tInputCommandMap = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		--["ADAPTER PORT"] = "33", 
		["HDMI 1"] 	= 'input "hdmi1"',
		["HDMI 2"] 	= 'input "hdmi2"',
		["HDMI 3"] 	= 'input "hdmi3"',
		["HDMI 4"] 	= 'input "hdmi4"',
		["DP 1"] 		= 'input "dp1"',
		["DP 2"] 		= 'input "dp2"',
		["DP 3"] 		= 'input "dp3"',
		["DP 4"] 		= 'input "dp4"',
	}
	
	tInputResponseMap = ReverseTable(tInputCommandMap)	-- Reverses the tInputCommandMap table	
	
	tInputCommandMap_Serial = {
		--index:  Connection Name
		--value:  Protocol Command Data	
		--["ADAPTER PORT"] = "33", 
		["HDMI 1"] = tohex('00 01 00 00 02'),
		["HDMI 2"] = tohex('00 01 00 00 03'),
	}
		
	tInputCommandMap_IR = { 
		--index:  Connection Name
		--value:  IR Code ID
		--["ADAPTER PORT"] = "51362",
		[""] = "",	
	}

	----------------------------------------- [*I/O HELPER TABLES*] -----------------------------------------
	tOutputConnMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Output Connection Name
		[0] = "Main Output",
	}

	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		--[0] = {Name = "INPUT HDMI 1",BindingID = PROJECTOR_PROXY_BINDINGID,},
		--The first two entries below are for the antenna connections which are bound to the tuner proxy; the in values will need to be updated base on the connection ID.
		[0] = {Name = "HDMI 1",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[1] = {Name = "HDMI 2",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[2] = {Name = "HDMI 3",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[3] = {Name = "HDMI 4",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[4] = {Name = "DP 1",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[5] = {Name = "DP 2",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[6] = {Name = "DP 3",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[7] = {Name = "DP 4",BindingID = PROJECTOR_PROXY_BINDINGID,},
	}
	
	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--["INPUT HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 2"] = {ID = 1,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 3"] = {ID = 2,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 4"] = {ID = 3,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["DP 1"] = {ID = 4,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["DP 2"] = {ID = 5,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["DP 3"] = {ID = 6,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["DP 4"] = {ID = 7,BindingID = PROJECTOR_PROXY_BINDINGID,},
	}

end	

function PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()")
	
end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

 end)
package.preload['device_messages'] = (function (...)
--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    Global Delarations
===============================================================================]]
gStatusError = {
    ["NoError"]	= tohex("00 00"),
    ["LampError"]	= tohex("00 01"),
    ["FanError"]	= tohex("00 02"),
    ["CoverError"]	= tohex("00 04"),
    ["TempError"]	= tohex("00 08"),
    ["D5VError"]	= tohex("00 10"),
    ["PowerError"]	= tohex("00 20"),
    ["TempWarning"]	= tohex("00 40"),
}

gStatusError_Network = {
    ["NoError"]	= '"no_err"',
    ["LampError"]	= '"err_light_src"',
    ["FanError"]	= '"err_fan"',
    ["CoverError"]	= '"err_cover"',
    ["TempError"]	= '"err_temp"',
    ["PowerError"]	= '"err_power"',
    ["PowerError2"]	= '"err_power2"',
    ["TempWarning"]	= '"err_temp"',
    ["DropShock"]	= '"err_shock"',
    ["NoLens"]	= '"err_nolens"',
    ["WheelError"]	= '"err_wheel"',
    ["LuminanceError"]	= '"err_light_over"',
}

gStatusError2 = {
    ["NoError"]	= tohex("00 00"),
    ["HighlandWarning"]	= tohex("00 20"),
}

gStatusPower = {
    ["Standby"]	= tohex("00 00"),
    ["StartUp"]	= tohex("00 01"),
    ["StartupLamp"]	= tohex("00 02"),
    ["PowerOn"]	= tohex("00 03"),
    ["Cooling1"]	= tohex("00 04"),
    ["Cooling2"]	= tohex("00 05"),
}

gStatusPower_Network = {
    ["Standby"]	= '"standby"',
    ["StartUp"]	= '"startup"',
    ["StartupLamp"]	= '',
    ["PowerOn"]	= '"on"',
    ["Cooling1"]	= '"cooling1"',
    ["Cooling2"]	= '"cooling2"',
}

gStatusInput = {
    ["HDMI1"]	= tohex("00 02"),
    ["HDMI2"]	= tohex("00 03"),
}

gStatusInput_Network = {
    ["HDMI1"]	= '"hdmi1"',
    ["HDMI2"]	= '"hdmi2"',
}

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
    if (gReceiveBuffer:len() == 0) then return "" end
    
    local message, pos, etx
    local pattern = ""
    if (gControlMethod == "Serial") then	
	   etx = string.pack("b", 0x9A)
	   pattern = "(.-" .. etx .. ")()"
    else	--ADCP Network
	   etx = "\r\n"
	   pattern = "(.-)" .. etx .. "()"
    end
    
    message, pos = string.match(gReceiveBuffer, pattern)
    
    if (message == nil) then
	   --LOG:Info("Do not have a complete message")
	   return ""
    end
    
    gReceiveBuffer = gReceiveBuffer:sub(pos)		

    return message

	
end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)

    if (type(message) == "table") then
	   LogTrace("HandleMessage. Message is ==>")
	   LogTrace(message)
	   LogTrace("<==")
    else
	   if (gControlMethod == "Network") then
		  LogTrace("HandleMessage. Message is ==>%s<==", message)
	   else
		  LogTrace("HandleMessage:" )
		  dbgHexdump(message)
	   end
    end
    
    --for NOKEY report during netowrk control initialization...
    if (gCon._LastCommand == nil) then return end
    
    local last_command = gCon._LastCommand.command_name
    LogInfo('last_command: ' .. tostring(last_command))
    
    local name, value, request, item, msg_type, data, data_len, pos, pattern, len
    local bAdvanceQueueOnly = false
    local bACK = false
    
    if (gControlMethod == "Network") then
	   if (message == "ok") then
		  bACK = true
	   elseif (message == "NOKEY") then
		  --bAdvanceQueueOnly = true
		  bACK = true
	   else
		  if (last_command == "SendKeepAlivePollingCommand") then
			 bAdvanceQueueOnly = true
		  elseif (last_command == "GetDevicePowerStatus") then
			 name = "POWER"
			 value = message
		  elseif (last_command == "GetDeviceInputStatus") then
			 name = "INPUT"
			 value = message
		  elseif (last_command == "GetDeviceVolumeStatus: Volume") then
			 name = "VOLUME"
			 value = message
		  elseif (last_command == "GetDeviceVolumeStatus: Mute") then
			 name = "MUTE"
			 value = message
		  elseif (last_command == "GetLampHours") then
			 name = "LAMP"
			 value = get_value_from_json_data(message, "light_src")
		  else
			 bAdvanceQueueOnly = true
		  end

	   end
	   
    else	--Serial
    
	   local stx_simplified_command = string.pack("b", 0xA9)
	   pattern =  stx_simplified_command .. "(..)(.)(..)"
	   item, msg_type, data = string.match(message, pattern)
	   value = data
	   
	   if (item == tohex("00 00")) then
		  bACK = true
	   elseif (item == tohex("00 01")) then 
		  name = "INPUT"
	   elseif (item == tohex("01 01")) then 
		  if (msg_type == tohex("03")) then 
			 LogWarn("NAK: Undefined Command")
			 bAdvanceQueueOnly = true
		  else	--msg_type == tohex("02")
			 name = "ERROR"
		  end
	   elseif (item == tohex("01 04")) then
			 LogWarn("NAK: Size Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("01 05")) then
			 LogWarn("NAK: Select Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("01 06")) then
			 LogWarn("NAK: Range Over")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("01 0A")) then
			 LogWarn("NAK: Not Applicable")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("01 02")) then 
		  name = "POWER"
	   elseif (item == tohex("01 13")) then 
		  name = "LAMP"
	   elseif (item == tohex("01 25")) then 
		  name = "ERROR2"
	   elseif (item == tohex("90 00")) then 
		   name = "MAC_ADDRESS"
	   elseif (item == tohex("90 01")) then 
		   name = "IP_ADDRESS"
	   elseif (item == tohex("F0 10")) then
			 LogWarn("NAK: Check Sum Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("F0 20")) then
			 LogWarn("NAK: Framing Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("F0 30")) then
			 LogWarn("NAK: Parity Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("F0 40")) then
			 LogWarn("NAK: Over Run Error")
			 bAdvanceQueueOnly = true
	   elseif (item == tohex("F0 50")) then
			 LogWarn("NAK: Other Comm Error")
			 bAdvanceQueueOnly = true
	   end
    
    end
    
    if (bAdvanceQueueOnly or bACK) then
	   advance_command_queue()
	   
	   if bACK then
		  if (last_command == "ON" or last_command == "OFF" or message == "NOKEY") then
			 gQueryPowerDelayTimer:StartTimer(1, "SECONDS")
		  elseif (last_command == "SET_INPUT") then
			 GetDeviceInputStatus()
		  elseif (last_command:find("VOL")) then
			 GetDeviceVolumeStatus()
		  end
	   end
	   
	   return
    else
	   advance_command_queue()
    end

    --name = name or message
    --value = value or ""	

    --TODO is set up if then block to test hex values for name & then set name to a dev_msg string value....

    DispatchMessage(name, value)

end

function advance_command_queue()    
    gCon._CommandRetryCount = 0
    gCon._WaitResponseTimer:KillTimer()
    local command_delay
    local delay_units = "MILLISECONDS"
    
    if (gProjectorProxy._VolumeIsRamping == true) then
	   command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	   delay_units = "MILLISECONDS"
    elseif (string.find(gCon._LastCommand.command_name, "PicturePosition")) then
	   command_delay = 20
	   delay_units = "SECONDS"
	   LogInfo('Picture Position in progress, 20 second delay before next command will be sent...')
    else
	   command_delay = tonumber(Properties["Command Delay Milliseconds"])
	   delay_units = "MILLISECONDS"
    end
    gHandleACKDelayTimer:StartTimer(command_delay, delay_units)
end

function get_value_from_json_data(sData, name)
    local value = ""
    local tData = JSON:decode(sData)
    for j,k in pairs(tData) do 
	   if (j == name) then return k end
	   if (type(k) == "table") then
		  for l,m in pairs(k) do
			 if (l == name) then return m end
		  end
	   end
    end
    
    return value 

end

function OnHandleACKDelayTimerExpired()
    LogTrace("OnHandleACKDelayTimerExpired")
    gHandleACKDelayTimer:KillTimer()
    
    gCon:HandleACK()
end


function OnPowerOnTimerExpired()
    LogTrace("OnPowerOnTimerExpired")
    gPowerOnTimer:KillTimer()
    
    gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(2, "SECONDS")	
    
--[[
    if gSetInputDelayTimer:TimerStarted() then
		  gSetInputDelayTimer:StartTimer(4, "SECONDS")
    end
		  
    gQueryVolumeDelayTimer:StartTimer(8, "SECONDS")
    gQueryVolumeDelayTimer._CallbackParam = "Volume_Mute" 
--]]

end

function OnPowerOffTimerExpired()
    LogTrace("OnPowerOffTimerExpired")
    gPowerOffTimer:KillTimer()
    
    --when using WOL, the device will not respond with a STANDBY message, so the proxy notfifcation will happen here...
    if (gProjector_Operating_Status ~= "STANDBY") then
	   gProjector_Operating_Status = "STANDBY"
	   C4:SetVariable('Projector_Operating_Status',"STANDBY")
	   gProjectorProxy:dev_PowerOff()
    end
    
    --for quick power on after power off...
    if gPowerOnTimer:TimerStarted() then
	   gSendPowerOnAfterPowerOffDelayTimer:StartTimer(2, "SECONDS")
    end

end

function OnSendPowerOnAfterPowerOffDelayTimerExpired()
    LogTrace("OnSendPowerOnAfterPowerOffDelayTimerExpired")
    gSendPowerOnAfterPowerOffDelayTimer:KillTimer()
    --gCheckPowerStateFromOnCommand = false
    gPowerOnTimer:StartTimer(tonumber(Properties["Power On Delay Seconds"]), "SECONDS")    
    ON()
end

function OnQueryPowerDelayTimerExpired()
    LogTrace("OnQueryPowerDelayTimerExpired")
    gQueryPowerDelayTimer:KillTimer()
       
    GetDevicePowerStatus()
    
    if gPowerOnTimer:TimerStarted() then
	   gQueryPowerDelayTimer:StartTimer(2, "SECONDS")
    end
    
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		--LogInfo("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhanded command = " .. MsgKey)
	end
end

--[[
TODO: Create DEV_MSG functions for all messages to call Notifies.
	  Sample functions are included below for all applicable notifications.
--]]

function DEV_MSG.MAC_ADDRESS(value)
    local mac = HexToString(value)
    C4:UpdateProperty("MAC Address", mac)
    
    LogTrace("DEV_MSG.MAC_ADDRESS(), value = " .. mac)
end

function DEV_MSG.IP_ADDRESS(value)
    local _, b1, b2, b3, b4 = string.unpack(value, "bbbb")
    local ip = b1 .."." .. b2 .."." .. b3 .."." .. b4
    C4:UpdateProperty("IP Address", ip)
    
    LogTrace("DEV_MSG.IP_ADDRESS(), value = " .. ip)
end

function DEV_MSG.ERROR(value)
	local tStatusError
	if (gControlMethod == "Network") then
	   tStatusError = gStatusError_Network
     else
	   tStatusError = gStatusError
	end


	if (value == gStatusError.NoError) then
		if (gProjector_Error_Status ~= "NO ERROR") then
			gProjector_Error_Status = "NO ERROR"
			C4:SetVariable('Projector_Error_Status',"NO ERROR")
		end	
	elseif (value== gStatusError.LampError) then
		if (gProjector_Error_Status ~= "LAMP ERROR") then
			gProjector_Error_Status = "LAMP ERROR"
			C4:SetVariable('Projector_Error_Status',"LAMP ERROR")
		end	
	elseif (value== gStatusError.FanError) then
		if (gProjector_Error_Status ~= "FAN ERROR") then
			gProjector_Error_Status = "FAN ERROR"
			C4:SetVariable('Projector_Error_Status',"FAN ERROR")
		end	
	elseif (value== gStatusError.CoverError) then
		if (gProjector_Error_Status ~= "COVER ERROR") then
			gProjector_Error_Status = "COVER ERROR"
			C4:SetVariable('Projector_Error_Status',"COVER ERROR")
		end
	elseif (value== gStatusError.TempError) then
		if (gProjector_Error_Status ~= "TEMP ERROR") then
			gProjector_Error_Status = "TEMP ERROR"
			C4:SetVariable('Projector_Error_Status',"TEMP ERROR")
		end
	elseif (value== gStatusError.D5VError) then
		if (gProjector_Error_Status ~= "D5V ERROR") then
			gProjector_Error_Status = "D5V ERROR"
			C4:SetVariable('Projector_Error_Status',"D5V ERROR")
		end	
	elseif (value== gStatusError.PowerError) then
		if (gProjector_Error_Status ~= "POWER ERROR") then
			gProjector_Error_Status = "POWER ERROR"
			C4:SetVariable('Projector_Error_Status',"POWER ERROR")
		end	
	elseif (value== gStatusError.PowerError2) then
		if (gProjector_Error_Status ~= "POWER ERROR2") then
			gProjector_Error_Status = "POWER ERROR2"
			C4:SetVariable('Projector_Error_Status',"POWER ERROR2")
		end	
	elseif (value== gStatusError.TempWarning) then
		if (gProjector_Error_Status ~= "TEMP WARNING") then
			gProjector_Error_Status = "TEMP WARNING"
			C4:SetVariable('Projector_Error_Status',"TEMP WARNING")
		end	
	elseif (value== gStatusError.DropShock) then
		if (gProjector_Error_Status ~= "DROP SHOCK") then
			gProjector_Error_Status = "DROP SHOCK"
			C4:SetVariable('Projector_Error_Status',"DROP SHOCK")
		end	
	elseif (value== gStatusError.NoLens) then
		if (gProjector_Error_Status ~= "LENS NOT ATTACHED") then
			gProjector_Error_Status = "LENS NOT ATTACHED"
			C4:SetVariable('Projector_Error_Status',"LENS NOT ATTACHED")
		end		
	elseif (value== gStatusError.WheelError) then
		if (gProjector_Error_Status ~= "WHEEL ERROR") then
			gProjector_Error_Status = "WHEEL ERROR"
			C4:SetVariable('Projector_Error_Status',"WHEEL ERROR")
		end	
	elseif (value== gStatusError.LuminanceError) then
		if (gProjector_Error_Status ~= "LUMINANCE ERROR") then
			gProjector_Error_Status = "LUMINANCE ERROR"
			C4:SetVariable('Projector_Error_Status',"LUMINANCE ERROR")
		end	
     else
	   gProjector_Error_Status = "UNHANDLED ERROR STATUS: " .. tostring(value)
	end
	LogTrace("DEV_MSG.ERROR(), value = " .. gProjector_Error_Status)
end

function DEV_MSG.ERROR2(value)
	if (value == gStatusError2.NoError) then
		if (gProjector_Error2_Status ~= "NO ERROR") then
			gProjector_Error2_Status = "NO ERROR"
			C4:SetVariable('Projector_Error2_Status',"NO ERROR")
		end	
	elseif (value== gStatusError2.HighlandWarning) then
		if (gProjector_Error2_Status ~= "Highland Warning") then
			gProjector_Error2_Status = "Highland Warning"
			C4:SetVariable('Projector_Error2_Status',"Highland Warning")
		end		
     else
	   gProjector_Error2_Status = "UNKNOWN ERROR2 STATUS!"
	end
	LogTrace("DEV_MSG.ERROR2(), value = " .. gProjector_Error2_Status)
end

function DEV_MSG.INPUT(value)
	local input = 0
	local input_name = ""
	
	
	local tStatusInput
	if (gControlMethod == "Network") then
	   tStatusInput = gStatusInput_Network
     else
	   tStatusInput = gStatusInput
	end
	
	if (value == tStatusInput.HDMI1) then
	   input = tInputConnMapByName["HDMI 1"].ID 
	   input_name = "HDMI 1"
	elseif (value == tStatusInput.HDMI2) then
	   input = tInputConnMapByName["HDMI 2"].ID
	   input_name = "HDMI 2"
	else
	   LogTrace("DEV_MSG.INPUT(), Invalid value!")
	   return
	end
     local bindingID = PROJECTOR_PROXY_BINDINGID	
	LogTrace("INPUT_CHANGED, input = " .. input_name)
	
	--need to handle audio and video legs of the input...
	gProjectorProxy:dev_InputChanged(bindingID, input + 3000)		--audio
	gProjectorProxy:dev_InputChanged(bindingID, input + 1000)		--video
	
	
	-- NOTE: there is TV proxy bug that affects Current Input in Composer Programming   
	-- if connection has audio then it is considered audio, if not then it is considered video	
	-- for video only connections (i.e. HDMI), you will need to implement logic to only send the video leg.
end

function DEV_MSG.POWER(value)
     local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
     local delay_units = "SECONDS"
	
	local tStatusPower
	if (gControlMethod == "Network") then
	   tStatusPower = gStatusPower_Network
     else
	   tStatusPower = gStatusPower
	end
	
	if (value == tStatusPower.Standby) then
		if (gProjector_Operating_Status ~= "STANDBY") then
			gProjector_Operating_Status = "STANDBY"
			C4:SetVariable('Projector_Operating_Status',"STANDBY")
			gProjectorProxy:dev_PowerOff()
		end
	     gPowerOffTimer:StartTimer(1, "SECONDS")
	elseif (value== tStatusPower.StartUp) then
		if (gProjector_Operating_Status ~= "START UP") then
			gProjector_Operating_Status = "START UP"
			C4:SetVariable('Projector_Operating_Status',"START UP")
		end	
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== tStatusPower.StartupLamp) then	
		if (gProjector_Operating_Status ~= "STARTUP LAMP") then
			gProjector_Operating_Status = "STARTUP LAMP"
			C4:SetVariable('Projector_Operating_Status',"STARTUP LAMP")
		end
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== tStatusPower.PowerOn) then	
	
		gPowerOnTimer:KillTimer()
		gQueryPowerDelayTimer:KillTimer()
		
		if (gProjector_Operating_Status ~= "POWER ON") then
			gProjector_Operating_Status = "POWER ON"
			C4:SetVariable('Projector_Operating_Status',"POWER ON")
			gProjectorProxy:dev_PowerOn()
			
		     if gSetInputDelayTimer:TimerStarted() then
			    gSetInputDelayTimer:StartTimer(1, "SECONDS")
		     end	
			
			GetLampHours()
		end	
	elseif (value== tStatusPower.Cooling1) then	
		if (gProjector_Operating_Status ~= "COOLING1") then
			gProjector_Operating_Status = "COOLING1"
			C4:SetVariable('Projector_Operating_Status',"COOLING1")
		end
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== tStatusPower.Cooling2) then	
		if (gProjector_Operating_Status ~= "COOLING2") then
			gProjector_Operating_Status = "COOLING2"
			C4:SetVariable('Projector_Operating_Status',"COOLING2")
		end
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
     elseif (value== "err_internal1" or value== "err_internal2") then
	   LogWarn("DEV_MSG.POWER(), value = " .. value .. ". Sending Power Status Query again...")
	   gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)
	else
		gProjector_Operating_Status = "UNKNOWN OPERATING STATUS!"
	end		
	LogTrace("DEV_MSG.POWER(), value = " .. gProjector_Operating_Status)
end

function DEV_MSG.LAMP(value)
    local hours = ""
    if (gControlMethod == "Network") then
	   hours = value
    else
	   _, hours = string.unpack(value, ">h")
    end
    
    LogTrace("Projector reports " .. hours .. " Hours on the bulb")	
    C4:UpdateProperty('Lamp Hours', hours)
    if (gLamp_Hours ~= hours) then
	   gLamp_Hours = hours
	   C4:SetVariable('Lamp_Hours',tonumber(hours))
    end	
end

function DEV_MSG.VOLUME(value)
	LogTrace("DEV_MSG.VOLUME(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = "" --mod 1000 value of Output Connection ID
	
	-- TODO: edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local deviceLevel = value
	local minDeviceLevel = -80
	local maxDeviceLevel = 20
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)		
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.MUTE(value)
	LogTrace("DEV_MSG.MUTE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == "01") then 
		state = "True"
	elseif (value == "00") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	gProjectorProxy:dev_MuteChanged(output, state)
end		

function DEV_MSG.BASS(value)
	LogTrace("DEV_MSG.BASS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BassLevelChanged(output, level)
end	

function DEV_MSG.TREBLE(value)
	LogTrace("DEV_MSG.TREBLE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Treble level is represented as a percentage value
	local level   
	
	gProjectorProxy:dev_TrebleLevelChanged(output, level)
end	

function DEV_MSG.BALANCE(value)
	LogTrace("DEV_MSG.BALANCE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BalanceLevelChanged(output, level)
end	

function DEV_MSG.LOUDNESS(value)
	LogTrace("DEV_MSG.LOUDNESS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "state", Loudness state is represented as "True" or "False" (literal string, not boolean)
	local state  
	gProjectorProxy:dev_LoudnessChanged(output, state)
end
	
 end)
package.preload['device_specific_commands'] = (function (...)
--[[=============================================================================
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_specific_commands = "2015.03.31"
end

gPictureMode_Serial = {
    ["BRT Cinema"] 		= tohex("00 06"),
    ["BRT TV"] 			= tohex("00 07"),
    ["Cinema Film 1"] 	= tohex("00 00"),
    ["Cinema Film 2"] 	= tohex("00 01"),
    --["Cinema Digital"] 	= "",
    ["Game"] 			= tohex("00 05"),
    ["Photo"] 			= tohex("00 04"),
    ["Reference"] 		= tohex("00 02"),
    ["TV"]				= tohex("00 03"),
    ["User"] 			= tohex("00 08"),
    --["User 1"] 			= "",
    --["User 2"] 			= "",
    --["User 3"] 			= "",
}

gPictureMode_Network = {
    ["BRT Cinema"] 		= '"brt_cinema"',
    ["BRT TV"] 			= '"brt_tv"',
    ["Cinema Film 1"] 	= '"cinema_film1"',
    ["Cinema Film 2"] 	= '"cinema_film2"',
    ["Cinema Digital"] 	= '"cinema_digital"',
    ["Game"] 			= '"game"',
    ["Photo"] 			= '"photo"',
    ["Reference"] 		= '"reference"',
    ["TV"]				= '"tv"',
    ["User"] 			= '"user"',
    ["User 1"] 			= '"user1"',
    ["User 2"] 			= '"user2"',
    ["User 3"] 			= '"user3"',
}


function EX_CMD.PictureMode(tParams)
    local command_name = "PictureMode"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
	
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 02")
	   local cmd_type = tohex("00")	--SET
	   local data = gPictureMode_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'picture_mode ' .. gPictureMode_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

--name of function in legacy drivers...
EX_CMD.CalibPreset = EX_CMD.PictureMode

gColorTemp_Serial = {
    ["D93"] 		= tohex("00 00"),
    ["D75"] 		= tohex("00 01"),
    ["D65"] 		= tohex("00 02"),
    ["Custom1"] 	= tohex("00 03"),
    ["Custom2"] 	= tohex("00 04"),
    ["Custom3"] 	= tohex("00 05"),
    ["Custom4"] 	= tohex("00 06"),
    ["Custom5"] 	= tohex("00 08"),
    ["D55"] 		= tohex("00 09"),
}

gColorTemp_Network = {
    ["D93"] 		= '"d93"',
    ["D75"] 		= '"d75"',
    ["D65"] 		= '"d65"',
    ["Custom1"] 	= '"custom1"',
    ["Custom2"] 	= '"custom2"',
    ["Custom3"] 	= '"custom3"',
    ["Custom4"] 	= '"custom4"',
    ["Custom5"] 	= '"custom5"',
    ["D55"] 		= '"d55"',
    ["DCI"] 		= '"dci"',
    
}


function EX_CMD.ColorTemperature(tParams)
    local command_name = "ColorTemperature"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 17")
	   local cmd_type = tohex("00")	--SET
	   local data = gColorTemp_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'color_temp ' .. gColorTemp_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gLampControl_Serial = {
    ["Low"] 		= tohex("00 00"),
    ["High"] 		= tohex("00 01"),
}

gLampControl_Network = {
    ["Low"] 		= '"low"',
    ["High"] 		= '"high"',
}

function EX_CMD.LampControl(tParams)
    local command_name = "LampControl"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])

    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 1A")
	   local cmd_type = tohex("00")	--SET
	   local data = gLampControl_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'lamp_control ' .. gLampControl_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gContrastEnhancer_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Low"] 		= tohex("00 01"),
    ["Middle"] 	= tohex("00 02"),
    ["High"] 		= tohex("00 03"),
}

gContrastEnhancer_Network = {
    ["Off"] 		= '"off"',
    ["Low"] 		= '"low"',
    ["Middle"] 	= '"mid"',
    ["High"] 		= '"high"',
}

function EX_CMD.ContrastEnhancer(tParams)
    local command_name = "ContrastEnhancer"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 1C")
	   local cmd_type = tohex("00")	--SET
	   local data = gContrastEnhancer_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'contrast_enh ' .. gContrastEnhancer_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
end

gAdvancedIris_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Full"] 		= tohex("00 02"),
    ["Limited"] 	= tohex("00 03"),
}

gAdvancedIris_Network = {
    ["Off"] 		= '"off"',
    ["Full"] 		= '"full"',
    ["Limited"] 	= '"limited"',
}

function EX_CMD.AdvancedIris(tParams)
    local command_name = "AdvancedIris"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 1D")
	   local cmd_type = tohex("00")	--SET
	   local data = gAdvancedIris_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'iris_dyn_cont ' .. gAdvancedIris_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gFilmMode_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Auto"] 		= tohex("00 02"),
}

gFilmMode_Network = {
    ["Off"] 		= '"off"',
    ["Auto"] 		= '"auto"',
}

function EX_CMD.FilmMode(tParams)
    local command_name = "FilmMode"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 1F")
	   local cmd_type = tohex("00")	--SET
	   local data = gFilmMode_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'film_mode ' .. gFilmMode_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gGammaCorrection_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["1.8"] 		= tohex("00 01"),
    ["2.0"] 		= tohex("00 02"),
    ["2.1"] 		= tohex("00 03"),
    ["2.2"] 		= tohex("00 04"),
    ["2.4"] 		= tohex("00 05"),
    ["2.6"] 		= tohex("00 06"),
    ["Gamma7"] 	= tohex("00 07"),
    ["Gamma8"] 	= tohex("00 08"),
    ["Gamma9"] 	= tohex("00 09"),
    ["Gamma10"] 	= tohex("00 0A"),
}

gGammaCorrection_Network = {
    ["Off"] 		= '"off"',
    ["1.8"] 		= '"1.8"',
    ["2.0"] 		= '"2.0"',
    ["2.1"] 		= '"2.1"',
    ["2.2"] 		= '"2.2"',
    ["2.4"] 		= '"2.4"',
    ["2.6"] 		= '"2.6"',
    ["Gamma7"] 	= '"gamma7"',
    ["Gamma8"] 	= '"gamma8"',
    ["Gamma9"] 	= '"gamma9"',
    ["Gamma10"] 	= '"gamma10"',
}

function EX_CMD.GammaCorrection(tParams)
    local command_name = "GammaCorrection"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 22")
	   local cmd_type = tohex("00")	--SET
	   local data = gGammaCorrection_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'gamma_correction ' .. gGammaCorrection_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gNR_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Low"] 		= tohex("00 01"),
    ["Middle"] 	= tohex("00 02"),
    ["High"] 		= tohex("00 03"),
    ["Auto"] 		= tohex("00 04"),
}

gNR_Network = {
    ["Off"] 		= '"off"',
    ["Low"] 		= '"low"',
    ["Middle"] 	= '"mid"',
    ["High"] 		= '"high"',
    ["Auto"] 		= '"auto"',
}

function EX_CMD.NR(tParams)
    local command_name = "NR"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 25")
	   local cmd_type = tohex("00")	--SET
	   local data = gNR_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'nr ' .. gNR_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gPictureMuting_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["On"] 		= tohex("00 01"),

}

gPictureMuting_Network = {
    ["Off"] 		= '"off"',
    ["On"] 		= '"on"',
}

function EX_CMD.PictureMuting(tParams)
    local command_name = "PictureMuting"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 30")
	   local cmd_type = tohex("00")	--SET
	   local data = gPictureMuting_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'blank ' .. gPictureMuting_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gColorSpace_Serial = {
    ["BT.709"] 		= tohex("00 00"),
    ["Color Space1"] 	= tohex("00 03"),
    ["Color Space2"] 	= tohex("00 04"),
    ["Color Space3"] 	= tohex("00 05"),
    ["Custom"] 		= tohex("00 06"),
    --["Custom1"] 		= "",
    --["Custom2"] 		= "",
    ["BT.2020"] 		= tohex("00 08"),
    --["Adobe RGB"] 		= "",
    --["DCI"] 			= "",
}

gColorSpace_Network = {
    ["BT.709"] 		= '"bt709"',
    ["Color Space1"] 	= '"color_space1"',
    ["Color Space2"] 	= '"color_space2"',
    ["Color Space3"] 	= '"color_space3"',
    ["Custom"] 		= '"custom"',
    ["Custom1"] 		= '"custom1"',
    ["Custom2"] 		= '"custom2"',
    ["BT.2020"] 		= '"bt2020"',
    ["Adobe RGB"] 		= '"adobe_rgb"',
    ["DCI"] 			= '"dci"',
}

function EX_CMD.ColorSpace(tParams)
    local command_name = "ColorSpace"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 3B")
	   local cmd_type = tohex("00")	--SET
	   local data = gColorSpace_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'color_space ' .. gColorSpace_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gMotionflow_Serial = {
    ["Off"] 			= tohex("00 00"),
    ["Smooth High"] 	= tohex("00 01"),
    ["Smooth Low"] 		= tohex("00 02"),
    ["Impulse"] 		= tohex("00 03"),
    ["Combination"] 	= tohex("00 04"),
    ["True Cinema"] 	= tohex("00 05"),
}

gMotionflow_Network = {
    ["Off"] 			= '"off"',
    ["Smooth High"] 	= '"smooth_high"',
    ["Smooth Low"] 		= '"smooth_low"',
    ["Impulse"] 		= '"impulse"',
    ["Combination"] 	= '"combination"',
    ["True Cinema"] 	= '"true_cinema"',
}

function EX_CMD.Motionflow(tParams)
    local command_name = "Motionflow"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 59")
	   local cmd_type = tohex("00")	--SET
	   local data = gMotionflow_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'motionflow ' .. gMotionflow_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gxvColor_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["On"] 		= tohex("00 01"),

}

gxvColor_Network = {
    ["Off"] 		= '"off"',
    ["On"] 		= '"on"',
}

function EX_CMD.xvColor(tParams)
    local command_name = "xvColor"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 5A")
	   local cmd_type = tohex("00")	--SET
	   local data = gxvColor_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'xvcolor ' .. gxvColor_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gRealityCreation_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["On"] 		= tohex("00 01"),

}

gRealityCreation_Network = {
    ["Off"] 		= '"off"',
    ["On"] 		= '"on"',
}

function EX_CMD.RealityCreation(tParams)
    local command_name = "RealityCreation"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 67")
	   local cmd_type = tohex("00")	--SET
	   local data = gRealityCreation_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'real_cre ' .. gRealityCreation_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gColorCorrection_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["On"] 		= tohex("00 01"),

}

gColorCorrection_Network = {
    ["Off"] 		= '"off"',
    ["On"] 		= '"on"',
}

function EX_CMD.ColorCorrection(tParams)
    local command_name = "ColorCorrection"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 86")
	   local cmd_type = tohex("00")	--SET
	   local data = gColorCorrection_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'col_correction ' .. gColorCorrection_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gClearWhite_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Low"] 		= tohex("00 01"),
    ["High"] 		= tohex("00 02"),

}

gClearWhite_Network = {
    ["Off"] 		= '"off"',
    ["Low"] 		= '"low"',
    ["High"] 		= '"high"',
}

function EX_CMD.ClearWhite(tParams)
    local command_name = "ClearWhite"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 6B")
	   local cmd_type = tohex("00")	--SET
	   local data = gClearWhite_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'clear_white ' .. gClearWhite_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gMPEGNR_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Low"] 		= tohex("00 01"),
    ["Middle"] 	= tohex("00 02"),
    ["High"] 		= tohex("00 03"),
    ["Auto"] 		= tohex("00 04"),
}

gMPEGNR_Network = {
    ["Off"] 		= '"off"',
    ["Low"] 		= '"low"',
    ["Middle"] 	= '"mid"',
    ["High"] 		= '"high"',
    ["Auto"] 		= '"auto"',
}

function EX_CMD.MPEGNR(tParams)
    local command_name = "MPEGNR"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 6C")
	   local cmd_type = tohex("00")	--SET
	   local data = gMPEGNR_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'mnr ' .. gMPEGNR_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSmoothGradation_Serial = {
    ["Off"] 		= tohex("00 00"),
    ["Low"] 		= tohex("00 01"),
    ["Middle"] 	= tohex("00 02"),
    ["High"] 		= tohex("00 03"),
}

gSmoothGradation_Network = {
    ["Off"] 		= '"off"',
    ["Low"] 		= '"low"',
    ["Middle"] 	= '"mid"',
    ["High"] 		= '"high"',
}

function EX_CMD.SmoothGradation(tParams)
    local command_name = "SmoothGradation"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 6D")
	   local cmd_type = tohex("00")	--SET
	   local data = gSmoothGradation_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'smooth_grd ' .. gSmoothGradation_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gRealityCreationDatabase_Serial = {
    ["Mastered in 4K"] 	= tohex("00 00"),
    ["Normal"] 		= tohex("00 01"),
}

gRealityCreationDatabase_Network = {
    ["Mastered in 4K"] 	= '"mi4k"',
    ["Normal"] 		= '"normal"',
}

function EX_CMD.RealityCreationDatabase(tParams)
    local command_name = "RealityCreationDatabase"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 75")
	   local cmd_type = tohex("00")	--SET
	   local data = gRealityCreationDatabase_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'real_cre_db ' .. gRealityCreationDatabase_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gPicturePosition_Serial = {
    ["1.85:1"] 		= tohex("00 00"),
    ["2.35:1"] 		= tohex("00 01"),
    ["Custom1"] 		= tohex("00 02"),
    ["Custom2"] 		= tohex("00 03"),
    ["Custom3"] 		= tohex("00 04"),
}

gPicturePosition_Network = {
    ["1.85:1"] 		= '"1.85_1"',
    ["2.35:1"] 		= '"2.35_1"',
    ["Custom1"] 		= '"custom1"',
    ["Custom2"] 		= '"custom2"',
    ["Custom3"] 		= '"custom3"',
}

function EX_CMD.PicturePosition(tParams)
    local command_name = "PicturePosition"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 66")
	   local cmd_type = tohex("00")	--SET
	   local data = gPicturePosition_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'pic_pos_sel ' .. gPicturePosition_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gAspect_Serial = {
    ["Normal"] 		= tohex("00 01"),
    ["V Stretch"] 		= tohex("00 0B"),
    ["1.85:1 Zoom"] 	= tohex("00 0C"),
    ["2.35:1 Zoom"] 	= tohex("00 0D"),
    ["Stretch"] 		= tohex("00 0E"),
    ["Squeeze"] 		= tohex("00 0F"),
}

gAspect_Network = {
    ["Normal"] 		= '"normal"',
    ["V Stretch"] 		= '"v_stretch"',
    ["1.85:1 Zoom"] 	= '"1.85_1_zoom"',
    ["2.35:1 Zoom"] 	= '"2.35_1_zoom"',
    ["Stretch"] 		= '"stretch"',
    ["Squeeze"] 		= '"squeeze"',
}

function EX_CMD.Aspect(tParams)
    local command_name = "Aspect"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 20")
	   local cmd_type = tohex("00")	--SET
	   local data = gAspect_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'aspect ' .. gAspect_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gLensShift_Serial = {
    ["Up"] 		= tohex("17 72"),
    ["Down"] 		= tohex("17 73"),
    ["Left"] 		= tohex("19 02"),
    ["Right"] 		= tohex("19 03"),
}

gLensShift_Network = {
    ["Up"] 		= '"lens_shift_up"',
    ["Down"] 		= '"lens_shift_down"',
    ["Left"] 		= '"lens_shift_left"',
    ["Right"] 		= '"lens_shift_right"',
}

function EX_CMD.LensShift(tParams)
    local command_name = "LensShift"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    --infrared remote command syntax
    if (gControlMethod == "Serial") then
	   local item = gLensShift_Serial[tParams["Value"]]
	   local cmd_type = tohex("00")	--SET
	   local data = tohex("00 00")
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'key ' .. gLensShift_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gLensFocus_Serial = {
    ["Far"] 		= tohex("17 74"),
    ["Near"] 		= tohex("17 75"),
}

gLensFocus_Network = {
    ["Far"] 		= '"lens_focus_far"',
    ["Near"] 		= '"lens_focus_near"',
}

function EX_CMD.LensFocus(tParams)
    local command_name = "LensFocus"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])

    local command
    --infrared remote command syntax
    if (gControlMethod == "Serial") then
	   local item = gLensFocus_Serial[tParams["Value"]]
	   local cmd_type = tohex("00")	--SET
	   local data = tohex("00 00")
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'key ' .. gLensFocus_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gLensZoom_Serial = {
    ["Large"] 		= tohex("17 77"),
    ["Small"] 		= tohex("17 78"),
}

gLensZoom_Network = {
    ["Large"] 		= '"lens_zoom_up"',
    ["Small"] 		= '"lens_zoom_down"',
}

function EX_CMD.LensZoom(tParams)
    local command_name = "LensZoom"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])

    local command
    --infrared remote command syntax
    if (gControlMethod == "Serial") then
	   local item = gLensZoom_Serial[tParams["Value"]]
	   local cmd_type = tohex("00")	--SET
	   local data = tohex("00 00")
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'key ' .. gLensZoom_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSelect2D3DDisplay_Serial = {
    ["Auto"] 		= tohex("00 00"),
    ["2D"] 		= tohex("00 01"),
    ["3D"] 		= tohex("00 02"),
}

gSelect2D3DDisplay_Network = {
    ["Auto"] 		= '"auto"',
    ["2D"] 		= '"2d"',
    ["3D"] 		= '"3d"',
}

function EX_CMD.Select2D3DDisplay(tParams)
    local command_name = "Select2D3DDisplay"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 60")
	   local cmd_type = tohex("00")	--SET
	   local data = gSelect2D3DDisplay_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = '2d3d_sel ' .. gSelect2D3DDisplay_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSelect3DFormat_Serial = {
    ["Simulated 3D"] 		= tohex("00 00"),
    ["Side-by-Side"] 		= tohex("00 01"),
    ["Over-Under"] 			= tohex("00 02"),
}

gSelect3DFormat_Network = {
    ["Simulated 3D"] 		= '"sim3d"',
    ["Side-by-Side"] 		= '"sidebyside"',
    ["Over-Under"] 			= '"overunder"',
}

function EX_CMD.Select3DFormat(tParams)
    local command_name = "Select3DFormat"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 61")
	   local cmd_type = tohex("00")	--SET
	   local data = gSelect3DFormat_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = '3d_format ' .. gSelect3DFormat_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSelect3DDepth_Serial = {
    ["-2"] 		= tohex("FF FE"),
    ["-1"] 		= tohex("FF FF"),
    ["0"] 		= tohex("00 00"),
    ["1"] 		= tohex("00 01"),
    ["2"] 		= tohex("00 02"),
}

gSelect3DDepth_Network = {
    ["-2"] 		= '-2',
    ["-1"] 		= '-1',
    ["0"] 		= '0',
    ["1"] 		= '+1',
    ["2"] 		= '+2',
}

function EX_CMD.Select3DDepth(tParams)
    local command_name = "Select3DDepth"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 62")
	   local cmd_type = tohex("00")	--SET
	   local data = gSelect3DDepth_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = '3d_depth ' .. gSelect3DDepth_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSelect3DSimulatedEffect_Serial = {
    ["High"] 		= tohex("00 00"),
    ["Middle"] 	= tohex("00 01"),
    ["Low"] 		= tohex("00 02"),
}

gSelect3DSimulatedEffect_Network = {
    ["High"] 		= '"high"',
    ["Middle"] 	= '"mid"',
    ["Low"] 		= '"low"',

}

function EX_CMD.Select3DSimulatedEffect(tParams)
    local command_name = "Select3DSimulatedEffect"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 63")
	   local cmd_type = tohex("00")	--SET
	   local data = gSelect3DSimulatedEffect_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'sim3d_effect ' .. gSelect3DSimulatedEffect_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gSelect3DBrightness_Serial = {
    ["High"] 		= tohex("00 00"),
    ["Standard"] 	= tohex("00 01"),
}

gSelect3DBrightness_Network = {
    ["High"] 		= '"high"',
    ["Standard"] 	= '"standard"',
}

function EX_CMD.Select3DBrightness(tParams)
    local command_name = "Select3DBrightness"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 72")
	   local cmd_type = tohex("00")	--SET
	   local data = gSelect3DBrightness_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = '3d_brt ' .. gSelect3DBrightness_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gHDR_Serial = {
    ["Off"] 			= tohex("00 00"),
    ["On"] 			= tohex("00 01"),
    ["Auto"] 			= tohex("00 02"),
    --["HDR10"] 			= "",
    --["HDR Reference"] 	= "",
    --["HLG"] 			= "",
}

gHDR_Network = {
    ["Off"] 			= '"off"',
    ["On"] 			= '"on"',
    ["Auto"] 			= '"auto"',
    ["HDR10"] 			= '"hdr10"',
    ["HDR Reference"] 	= '"hdr_reference"',
    ["HLG"] 			= '"hlg"',
    
}

function EX_CMD.HDR(tParams)
    local command_name = "HDR"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    
    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 7C")
	   local cmd_type = tohex("00")	--SET
	   local data = gHDR_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'hdr ' .. gHDR_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end



gAutoIris_Serial = {
	["Off"] 		= tohex("00 00"),
    ["Full"] 		= tohex("00 02"),
    ["Limited"] 	= tohex("00 03"),
}

gAutoIris_Network = {
	["Off"] 		= '"off"',
    ["Full"] 		= '"full"',
    ["Limited"] 	= '"limited"',
}

function EX_CMD.AutoIris(tParams)
    local command_name = "Auto Iris"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])

    local command
    if (gControlMethod == "Serial") then
	   local item = tohex("00 1D")
	   local cmd_type = tohex("00")	--SET
	   local data = gAutoIris_Serial[tParams["Value"]]
	   if (data ~= nil) then
		  command = item .. cmd_type .. data
	   end
    else	--Network
	   command = 'iris_dyn_cont ' .. gAutoIris_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end

gAdvanvedIrisLaserLightControl_Network = {
	["Off"] 		= '"off"',
    ["Full"] 		= '"full"',
    ["Limited"] 	= '"limited"',
}

function EX_CMD.AdvancedIrisandLaserLightControl(tParams)
    local command_name = "Advanced Iris and Laser Light Control"
    LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])

    local command
    if (gControlMethod == "Network") then
	   command = 'light_output_iris_dyn ' .. gAdvanvedIrisLaserLightControl_Network[tParams["Value"]]
    end
    
    if (command ~= nil) then
	   PackAndQueueCommand(command_name .. ": " .. tParams["Value"], command)
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"])
    else
	   LogTrace("EX_CMD." .. command_name .. ": " .. tParams["Value"] .. " is not avaialble with " .. gControlMethod .. " control.")
    end
    
end
 end)
package.preload['actions'] = (function (...)
--[[=============================================================================
    Lua Action Code

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.actions = "2016.01.08"
end

-- TODO: Create a function for each action defined in the driver

function LUA_ACTION.TemplateVersion()
	TemplateVersion()
end


function LUA_ACTION.PrintVolumeCurve()
    print("===== Volume Curve =====")
    for j,k in pairs(tVolumeCurve) do
	   print(j,k)
    end
end

function LUA_ACTION.GetMACandIPAddresses()
    GetMACAddress()
    GetIPAddress()
end

 end)
package.preload['common.c4_utils'] = (function (...)
--[[=============================================================================
    Helper functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_utils = "2016.01.08"
end

--[[=============================================================================
    AsciiToBCD(InString)

    Description
    Convert an ascii string to a binary coded decimal. Each decimal digit is
    stored in one byte, with the lower four bits encoding the digit in BCD form.

    Parameters
    InString(string) - Ascii string that is to be converted into bcd

    Returns
    The binary coded decimal
===============================================================================]]
function AsciiToBCD(InString)
	local WorkVal = 0
	local RetValStr = ""
	local DoingHighNybble = false
	local WorkStr = ((#InString % 2) == 0) and (InString) or ("0" .. InString)	-- make sure length is an even number

	for CharCount = 1, #WorkStr do
		local NumVal = tonumber(WorkStr:sub(CharCount, CharCount))

		WorkVal = bit.lshift(WorkVal, 4) + NumVal
		if (DoingHighNybble) then
			RetValStr = RetValStr .. string.char(WorkVal)
			WorkVal = 0
		end

		DoingHighNybble = (not DoingHighNybble)
	end

	return RetValStr
end

--[[=============================================================================
    BCDToAscii(InByte)

    Description
    Convert an BCD string to an ascii string.

    Parameters
    InByte(string) - Binary coded decimal that is to be converted into ascii

    Returns
    The ascii string
===============================================================================]]
function BCDToAscii(InByte)
	return tostring(bit.rshift(InByte, 4)) .. tostring(bit.band(InByte, 0x0F))
end

--[[=============================================================================
    MakeXMLNode(Tag, Value)

    Description
    Create an Xml element

    Parameters
    Tag(string)   - The Xml elements name
    Value(string) - The Xml elements value

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLNode(Tag, Value)
	return "<" .. Tag .. ">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)

    Description
    Create an Xml element with an attribute

    Parameters
    Tag(string)       - The Xml elements name
    Value(string)     - The Xml elements value
    Attribute(string) - The attribute to be added to the Xml element
    AttrValue(string) - The value of the attribute to be added

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)
    return "<" .. Tag .. " " .. Attribute .. "=\"" .. AttrValue .. "\">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    StringFromUnicode(UnicodeString)

    Description
    Convert a unicode string

    Parameters
    UnicodeString(string) - The unicode string to be converted to ascii

    Returns
    The ascii representation of the unicode string
===============================================================================]]
function StringFromUnicode(UnicodeString)
	local RetVal = ""

	-- extract every other byte from the unicode string
	for Index = 2, #UnicodeString, 2 do
		RetVal = RetVal .. string.sub(UnicodeString, Index, Index)
	end

	return RetVal
end

--[[=============================================================================
    StringSplit(s)

    Description
    Splits a string into multiple strings at an optionally specified delimiter
	If the delimiter is not specified, it will defalt to the space character

    Parameters
    s(string) - The string that is to be split into several strings
	d(string) - The delimiter to split the string on

    Returns
    A table of strings containing all the seperate values in the given string
===============================================================================]]
function StringSplit(s, d)
	local delim = (d ~= nil) and d or " "
	local result = {}

	if s == nil or s == "" then
		return result
	end

	for match in (s..delim):gmatch("(.-)"..delim) do
		table.insert(result, match)
	end

	return result
end

--[[=============================================================================
    toboolean(s)

    Description
    Returns a boolean representation of the given value

    Parameters
    val input value, may be of different types

    Returns
    The value true or false based on the given value
		If the value is of type string the return true if the first letter is "T" or "t" or if the string is "1"
		If the value is of type number the return true if the value is non-zero
		If the value is already a boolean, just return it.
===============================================================================]]
function toboolean(val)
	local rval = false;

	if type(val) == "string" and (string.lower(val) == 'true' or val == "1") then
		rval = true
	elseif type(val) == "number" and val ~= 0 then
		rval =  true
	elseif type(val) == "boolean" then
		rval = val
	end

	return rval
end

--[[=============================================================================
    tointeger(s)

    Description
    Force a number or a string representation of a number to be an integer

    Parameters
    val - A number or a string representation of a number

    Returns
    The the rounded off integer value.
===============================================================================]]
function tointeger(val)
	local nval = tonumber(val)
	return (nval >= 0) and math.floor(nval + 0.5) or math.ceil(nval - 0.5)
end


--[[=============================================================================
    Go(to, err, ...)

    Description
    Call a function with the given arguments if it exists or report the error

    Parameters
    to(string)  - The string to evaluate the boolean representation from
    err(string) - The error to report if the function does not exist
    ...         - Additional optional parameters for the function specified by
                  the "to" parameter

    Returns
    Nothing
===============================================================================]]
function Go(to, err, ...)
	if (type(to) == "function") then
		return to(...)
	else
		LogTrace(err)
	end
end

--[[=============================================================================
    IsEmpty(str)

    Description
    Identifies if the string given is nil or empty

    Parameters
    str(string) - The string to evaluate for the empty condition

    Returns
    True if the given value is empty, false otherwise
===============================================================================]]
function IsEmpty(str)
	return str == nil or str == ""
end

--[[=============================================================================
    ReverseTable(a)

    Description
    Reverse table entries (key=value, value=key)

    Parameters
    a(table) - The table to reverse

    Returns
    new reversed table
===============================================================================]]
function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function tonumber_loc(str, base)
  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

--[[=============================================================================
    HexToString(InString)

    Description
    Converts a string of Hex characters to a readable string of ASCII characters

    Parameters
    InString(string) - The string to be converted

    Returns
    A string showing the hex bytes of the InString
===============================================================================]]
function HexToString(InString)
	local RetVal = ""

	for Index = 1, #InString do
		RetVal = RetVal .. string.format("%02X ", InString:byte(Index))
	end
	return RetVal
end


--[[=============================================================================
    StringToHex(InString)

    Description
    Converts a string of ASCII characters to as string with the actual Hex bytes in them.
	Basically an array of hex bytes.

    Parameters
    InString(string) - The string to be converted

    Returns
    A string of hex bytes (really an array of hex values) 
===============================================================================]]
function StringToHex(InString)
	local RetVal = ""

	for HexByteString in string.gfind(InString, "%x%x") do
		RetVal = RetVal .. string.char(tonumber(HexByteString, 16))
	end
	return RetVal
end

function RecordHistory(severity, eventType, category, subcategory, description)
	C4:RecordHistory(severity, eventType, category, subcategory, description)
end

function RecordCriticalHistory(eventType, category, subcategory, description)
	RecordHistory("Critical", eventType, category, subcategory, description)
end

function RecordWarningHistory(eventType, category, subcategory, description)
	RecordHistory("Warning", eventType, category, subcategory, description)
end

function RecordInfoHistory(eventType, category, subcategory, description)
	RecordHistory("Info", eventType, category, subcategory, description)
end


 end)
package.preload['common.c4_ir_connection'] = (function (...)
--[[=============================================================================
    Base for an IR connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_ir_connection = "2016.01.08"
end

IRConnectionBase = inheritsFrom(DeviceConnectionBase)

function IRConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function IRConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("tSerConBase:Initialize")
	gControlMethod = "IR"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function IRConnectionBase:ControlMethod()
	return "IR"
end

function IRConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendIR(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("IR connection is not bound. Command not sent.")
	end
end

function IRConnectionBase:destruct()
	self.superClass():destruct()
	self._BindingID = nil
end end)
package.preload['common.c4_serial_connection'] = (function (...)
--[[=============================================================================
    Base for a serial connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_device_connection_base"
require "lib.c4_log"
require "common.c4_common"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_serial_connection = "2016.01.08"
end

SerialConnectionBase = inheritsFrom(DeviceConnectionBase)

function SerialConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function SerialConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "Serial"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function SerialConnectionBase:ControlMethod()
	return "Serial"
end

function SerialConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendToSerial(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("Not connected to serial. Command not sent.")
	end
end

function SerialConnectionBase:SendRaw(sData)
	C4:SendToSerial(self._BindingID, sData)
end


function SerialConnectionBase:ReceivedFromSerial(idBinding, sData)
	self:ReceivedFromCom(sData)
end

function SerialConnectionBase:destruct()
	self.superClass():destruct()
	self._BindingID = nil
end end)
package.preload['lib.c4_queue'] = (function (...)
--[[=============================================================================
    c4_queue Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_queue = "2016.01.08"
end

c4_queue = inheritsFrom(nil)

function c4_queue:construct()
	-- entry table
	self._et = {first = 0, last = -1}
	self._maxSize = 0	-- no size limit
	self._name = ""

	local mt = getmetatable(self)
	if (mt ~= nil) then
		mt.__tostring = self.__tostring
	end
end

function c4_queue:__tostring()
	local tOutputString = {}
	table.insert(tOutputString, "--- Queue ---")
	if (not IsEmpty(self._name)) then
		table.insert(tOutputString, "  name = " .. tostring(self._name))
	end
	table.insert(tOutputString, "  first = " .. tostring(self._et.first))
	table.insert(tOutputString, "  last = " .. tostring(self._et.last))
	table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
	table.insert(tOutputString, "  maximum size = " .. self._maxSize)
	table.insert(tOutputString, "  next value = " .. tostring(self:value()))
	return table.concat(tOutputString, "\n")
end


-- push a value on the queue
function c4_queue:push(value, ...)
	local numItems = self._et.last - self._et.first + 1

	if ((self._maxSize <= 0) or (numItems < self._maxSize) ) then
		local last = self._et.last + 1
		self._et.last = last
		
		local interval = select(1, ...)
		local units = select(2, ...)
		local command_name = select(3, ...)
		self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}		
		--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
	else
		-- if addToQueue == true then push value to queue
		if (self:OnMaxSizeReached()) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}
			--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		end
	end
end

function c4_queue:OnMaxSizeReached()
	--LogTrace ("Max Size Reached - clear queue and push value to the queue (default).")
	local addToQueue = true

	self:clear()
	return (addToQueue)
end

-- pop a value from the queue
function c4_queue:pop()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace("Queue:pop(), queue is empty")
		return ""
	end

	local value = self._et[first]
	self._et[first] = nil        -- to allow garbage collection
	self._et.first = first + 1
	--LogTrace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)

	return value
end

-- clear queue
function c4_queue:clear()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace ("Queue:clear(), queue is empty")
		return ""
	end

	self._et = {first = 0, last = -1}
	--LogTrace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
	--LogTrace (self._et)
end

-- return value of first item
function c4_queue:value()
	local first = self._et.first

	if (first > self._et.last) then
		return ""
	else
		return self._et[first]
	end
end

-- return queue's maximum size
function c4_queue:MaxSize()
	return self._maxSize
end

-- return queue's maximum size
function c4_queue:SetMaxSize(size)
	self._maxSize = size
end

function c4_queue:SetName(name)
	self._name = name
end

-- return the queue's current size
function c4_queue:Size()
	return self._et.last - self._et.first + 1
end

-- is queue empty?
function c4_queue:empty()
	-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
	if (self._et.first > self._et.last) then
		return true
	else
		return false
	end
end

--[[
    c4_queue unit tests
--]]
function __test_c4_queue()
	require "test.C4Virtual"
	require "lib.c4_log"

	local LOG = c4_log:new("test_c4_queue")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	-- create an instance of the queue
	local c4Queue = c4_queue:new()

	c4Queue:SetMaxSize(3)
	assert(c4Queue:MaxSize() == 3, "_maxSize is not equal to '3' it is: " .. c4Queue:MaxSize())

	c4Queue:push("Item #1 in Queue")
	c4Queue:push("Item #2 in Queue")
	c4Queue:push("Item #3 in Queue")
	c4Queue:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4Queue:Size() == 1, "queue size is not equal to '1' it is: " .. c4Queue:Size())

	print (c4Queue)

	-- Test inheritance overriding OnMaxSizeReached
	-- Create a new class
	c4_queue_new = inheritsFrom(c4_queue)

	-- override construct()
	function c4_queue_new:construct()
		self.superClass():construct()  -- call base class
		self._maxSizeOption = 1

		local mt = getmetatable(self)
		if (mt ~= nil) then
			mt.__tostring = self.__tostring
		end
	end

	-- override OnMaxSizeReached()
	function c4_queue_new:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,

		-- Option 1: Do Nothing, new item is not added to queue
		if (self._maxSizeOption == 1) then
			
			LogInfo("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)
		-- Option 2: pop value, and push new value on queue
		elseif(self._maxSizeOption == 2) then
			LogInfo("Max Size Reached - pop value, and push new value on queue (option 2)")
			self:pop()
			return (true)
		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(self._maxSizeOption == 3) then
			LogInfo("Max Size Reached - clear queue and DO NOT push new value onto queue")
			self:clear()
			return (false)
		end
	end

	-- create an instance of the new queue
	local c4QueueNew = c4_queue_new:new()
	c4QueueNew:SetMaxSize(3)
	c4QueueNew:push("Item #1 in Queue")
	c4QueueNew:push("Item #2 in Queue")
	c4QueueNew:push("Item #3 in Queue")
	c4QueueNew:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4QueueNew:Size() == 3, "queue size is not equal to '3' it is: " .. c4QueueNew:Size())

	print(c4QueueNew)
	print ("done...")
end end)
package.preload['common.c4_device_connection_base'] = (function (...)
--[[=============================================================================
    DeviceConnectionBase Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_common"
require "lib.c4_object"
require "lib.c4_log"
require "lib.c4_timer"
require "lib.c4_queue"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_device_connection_base = "2016.01.08"
end

COMMAND_QUEUE_SIZE = 100
DEFAULT_COMMAND_DELAY_INTERVAL = 100            -- Don't send consecutive commands faster than this many milliseconds
DEFAULT_COMMAND_RESPONSE_INTERVAL = 3           -- If we haven't received and ACK after this many seconds, try again
DEFAULT_RETRY_COUNT_MAX = 1

function ON_DRIVER_EARLY_INIT.c4_device_connection_base()
	gReceiveBuffer = ""
	gIsUrlConnected = false
	gIsNetworkServerConnected = false
	gIsNetworkConnected = false
	gIsSerialConnected = false
	gIsIRConnected = false
end

DeviceConnectionBase = inheritsFrom(nil)

function DeviceConnectionBase:construct()

	self._IsConnected = false
	self._SendTimer = nil
	self._WaitResponseTimer = nil
	self._CommandQueue = nil
	self._Priority1CommandQueue = nil
	self._Priority2CommandQueue = nil
	self._LastCommand = nil
	self._ExpectAck = false
	self._CommandRetryCount = 0
	self._RetryCountMax = DEFAULT_RETRY_COUNT_MAX

	self._SendCommandDelayMS = DEFAULT_COMMAND_DELAY_INTERVAL
	self._CommandResponseWaitS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	-- Polling
	self._PollingInterval = 0
	self._PollingUnits = "SECONDS"
	self._PollingTimer = nil
end

function DeviceConnectionBase:Initialize(ExpectAck, CommandDelayInterval, CommandResponseInterval, CallbackParam)

	if (ExpectAck ~= nil) then
		self._ExpectAck = ExpectAck
	end

	if (CommandDelayInterval ~= nil) then
		self._SendCommandDelayMS = CommandDelayInterval
	end

	if (CommandResponseInterval ~= nil) then
		self._CommandResponseWaitS = CommandResponseInterval
	end

	self._CommandQueue = c4_queue:new()
	self._CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._CommandQueue:SetName("Command Queue")

	self._Priority1CommandQueue = c4_queue:new()
	self._Priority1CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority1CommandQueue:SetName("P1 Queue")

	self._Priority2CommandQueue = c4_queue:new()
	self._Priority2CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority2CommandQueue:SetName("P2 Queue")

	-- usually only one of these timers will be used, but it's pretty low overhead to instantiate both of them
	self._SendTimer = c4_timer:new("SendCommand", self._SendCommandDelayMS, "MILLISECONDS", DeviceConnectionBase.OnSendTimeExpired, false, CallbackParam)
	self._WaitResponseTimer = c4_timer:new("WaitResponse", self._CommandResponseWaitS, "SECONDS", DeviceConnectionBase.OnWaitTimeExpired, false, CallbackParam)
end

function DeviceConnectionBase:InitPolling(PollingInterval, PollingUnits, CallbackParam)
	LogFatal("DeviceConnectionBase:InitPolling()")
	if (PollingInterval ~= nil) then
		self._PollingInterval = PollingInterval
	end
	
	self._PollingUnits = PollingUnits or self._PollingUnits

	LogFatal("self._PollingInterval: %s, self._PollingUnits: %s", tostring(self._PollingInterval), tostring(self._PollingUnits))
	
	-- create polling timer
	self._PollingTimer = c4_timer:new("Polling", self._PollingInterval, self._PollingUnits, DeviceConnectionBase.OnPollingTimerExpired, false, CallbackParam)
end

function DeviceConnectionBase:StartPolling(interval, units)
	LogFatal("DeviceConnectionBase:StartPolling()")
	LogFatal("self._PollingTimer: %s", tostring(self._PollingTimer))
	
	if (self._PollingTimer ~= nil) then
		self._PollingTimer:KillTimer()
		
		local timer_units = units or self._PollingTimer._units
		local timer_interval = interval or self._PollingInterval

		self._PollingTimer:StartTimer(timer_interval, timer_units)
	end
end

function DeviceConnectionBase:StopPolling()
	LogFatal("DeviceConnectionBase:StopPolling()")
	self._PollingTimer:KillTimer()
end

function DeviceConnectionBase:SetExpectACK(ExpectACK)
	self._ExpectAck = ExpectACK
end

function DeviceConnectionBase:SetCommandDelayInterval(DelayInterval)
	self._SendCommandDelayMS = DelayInterval
end

function DeviceConnectionBase:SetResponseWaitInterval(WaitInterval)
	self._CommandResponseWaitS = WaitInterval
end

function DeviceConnectionBase:ReceivedFromCom(sData)
    	
     if (gControlMethod == "Network") then
	   LogTrace("ReceivedFromCom, received sData is ==>%s<==", sData)
     else
	   LogTrace("ReceivedFromCom, received sData:")
	   dbgHexdump(sData)
     end	
     
	gReceiveBuffer = gReceiveBuffer .. sData
	LogTrace("ReceivedFromCom  ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)

	message = self:GetMessage()
	while (message ~= nil and message ~= "") do
		status, err = pcall(HandleMessage, message)
		if (status) then
			message = self:GetMessage()
		else
			LogError("LUA_ERROR: " .. err)
			message = ""
			gReceiveBuffer = ""
		end
	end
end

function DeviceConnectionBase:SetConnection(IsConnected, method)
	self._IsConnected = IsConnected
	gControlMethod = method
end

function DeviceConnectionBase:ControlMethod()
	-- Override in derived class
	print("WARNING: Need to override ControlMethod - should never be called")
	
	return ""
end

function DeviceConnectionBase:StartCommandTimer(...)
	local value = select(1, ...)
	local units = select(2, ...)
	local command_name = select(3, ...) or ""

	self._WaitResponseTimer:KillTimer()
	self._SendTimer:KillTimer()

	if (self._ExpectAck) then
		-- expecting an ACK set the Response Wait timer
		local timer_units = units or self._WaitResponseTimer._units
		local timer_interval = value or self._CommandResponseWaitS

		self._WaitResponseTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting wait Timer:  %d", self._WaitResponseTimer._timerID) .. " for " .. command_name)
	else
		-- no ACK expected, just wait the designated amount of time and send another command
		local timer_units = units or self._SendTimer._units
		local timer_interval = value or self._SendCommandDelayMS

		self._SendTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting Send Timer:  %d for %s (timer_interval = %d, timer_units = %s)", self._SendTimer._timerID, command_name, timer_interval, timer_units))
	end
end

-- Note the '.' instead of the ':'
function DeviceConnectionBase.OnSendTimeExpired(Instance)
	LogTrace("Send Timer expired")
	Instance._SendTimer:KillTimer()

	local tCommand = Instance._LastCommand
	if (tCommand ~= nil) then
		if (tCommand.command_name ~= nil) then
			LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand.command_name))
		elseif (type(tCommand) == "string") then
			LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand))
		end
	else
		LogTrace("Send Timer expired - Last Command: UNKNOWN, Send Next Command")
	end
	
	Instance._LastCommand = nil
	Instance:SendNextCommand()

	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase.OnWaitTimeExpired(Instance)
	LogTrace("Wait Timer expired")
	Instance._WaitResponseTimer:KillTimer()
	Instance._CommandRetryCount = Instance._CommandRetryCount + 1
	
	if (Instance._CommandRetryCount >= Instance._RetryCountMax) then
		-- To many retries, pop the current command and try the next one
		Instance._CommandRetryCount = 0
		Instance._LastCommand = nil
		Instance:SendNextCommand()
	else
		Instance:SendLastCommand()
	end
end

function DeviceConnectionBase.OnPollingTimerExpired(Instance)
	LogTrace("Polling Timer expired")
	Instance._PollingTimer:KillTimer()

	OnPollingTimerExpired()
	
	Instance._PollingTimer:StartTimer(Instance._PollingInterval)
end

function DeviceConnectionBase:HandleACK()
	self._LastCommand = nil
	
	self._WaitResponseTimer:KillTimer()
	self._CommandRetryCount = 0
	self:SendNextCommand()
	
	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase:QueueEmpty()
	return (self._CommandQueue:empty() and self._Priority1CommandQueue:empty() and self._Priority2CommandQueue:empty())
end

function DeviceConnectionBase:QueueCommand(sCommand, ...)
--	LogTrace("QueueCommand(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority1Command(sCommand, ...)
	LogTrace("QueuePriority1Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority1CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority2Command(sCommand, ...)
	LogTrace("QueuePriority2Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority2CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:SendNextCommand()
	LogTrace("DeviceConnectionBase:SendNextCommand")

	local tCommand = nil
	if (not self._Priority1CommandQueue:empty()) then
		tCommand = self._Priority1CommandQueue:pop()
		LogTrace(tostring(gCon._Priority1CommandQueue))
	elseif (not self._Priority2CommandQueue:empty()) then
		tCommand = self._Priority2CommandQueue:pop()
		LogTrace(tostring(gCon._Priority2CommandQueue))
	elseif (not self._CommandQueue:empty()) then
		tCommand = self._CommandQueue:pop()
		LogTrace(tostring(gCon._CommandQueue))
	end
	
	if (tCommand ~= nil) then
		self._LastCommand = tCommand
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
		     if (gControlMethod == "Network") then
			   LogTrace("SendCommand ==>%s<==", sCommand)
		     else
				LogTrace("SendCommand:")
				dbgHexdump(sCommand)
		     end	
			
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendLastCommand()
--	LogTrace("DeviceConnectionBase:SendLastCommand")

	local tCommand = self._LastCommand
	if (tCommand ~= nil) then
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendCommand()
	-- Dummy routine.  Override in derived class
	print("Need to override SendCommand - should never be called")
end


function DeviceConnectionBase:GetMessage()
	
	-- Brain dead version of this routine. Just return the current receive buffer.
	-- It's very likely that a GetMessage() function will need to be created
	if (GetMessage ~= nil and type(GetMessage) == "function") then
		return GetMessage()
	else
		local ComMessage = gReceiveBuffer
		gReceiveBuffer = ""

		return ComMessage
	end
end

function DeviceConnectionBase:destruct()

    if (self._CommandQueue ~= nil) then
	   self._CommandQueue:clear()
	   self._CommandQueue = nil
    end
    
	
	
    if (self._Priority1CommandQueue ~= nil) then
	   self._Priority1CommandQueue:clear()
	   self._Priority1CommandQueue = nil
    end
	
    if (self._Priority2CommandQueue ~= nil) then
	   self._Priority2CommandQueue:clear()
	   self._Priority2CommandQueue = nil
    end

	
    if (self._SendTimer ~= nil) then 
	   self._SendTimer:destruct() 
    end
			    
    if (self._WaitResponseTimer ~= nil) then 
	   self._WaitResponseTimer:destruct() 
    end
    
end

--[[=============================================================================
    Other Connection Functions
===============================================================================]]

function ReceivedFromSerial(idBinding, sData)
	if (gCon.ReceivedFromSerial == nil) then return end --serial is bound but not the current control method
	gCon:ReceivedFromSerial(idBinding, sData)
end

function ReceivedFromNetwork(idBinding, nPort, sData)
	gCon:ReceivedFromNetwork(idBinding, nPort, sData)
end

function OnServerDataIn(nHandle, strData)
--	LogTrace("Received Data on Handle: " .. nHandle .. ": " .. strData)
--	LogTrace("Data Is: %s", HexToString(strData))
	gCon:ReceivedFromNetworkServer(nHandle, strData)
end


--[[=============================================================================
    The ReceivedAsync function is called in response to 'url_get_request'. 
    The ticketId is the number returned from the request.
===============================================================================]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

--	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
--	LogTrace(tHeaders)

	gCon:ReceivedAsync(ticketId, strData, responseCode, tHeaders)
end
	
--[[=============================================================================
    OnBindingChanged(idBinding, class, bIsBound)
  
    Description:
    Function called by Director when a binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnBindingChanged(idBinding, class, bIsBound)
	
	LogTrace("OnBindingChanged(): idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	if (idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
		SetControlMethod()
		OnSerialConnectionChanged(idBinding, class, bIsBound)
	elseif (idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
		SetControlMethod()
		OnIRConnectionChanged(idBinding, class, bIsBound)
	elseif(OnConnectionChanged ~= nil and type(OnConnectionChanged) == "function") then
		OnConnectionChanged(idBinding, class, bIsBound)
	end
end

--[[=============================================================================
    OnNetworkBindingChanged(idBinding, bIsBound)
  
    Description:
    Function called by Director when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	LogTrace('OnNetworkBindingChanged(): idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))

	gIsNetworkConnected = bIsBound
	SetControlMethod()
	OnNetworkConnectionChanged(idBinding, bIsBound)
	if (bIsBound) then
		-- Start a special instance of reconnect timer to eventually do NetConnect if not done automatically
		gCon._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval) 	
	end	
end

--[[=============================================================================
    OnConnectionStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. tostring(nPort) .. ")]: " .. sStatus)

	local isOnline = false

	gNetworkStatus = sStatus	
	if (sStatus == "ONLINE") then
		isOnline = true
	end

	gCon:SetOnlineStatus(isOnline)
	OnNetworkStatusChanged(idBinding, nPort, sStatus)
end

--[[=============================================================================
    SetControlMethod()
  
    Description:
    Sets the control method type for the drivers internal infrastructure
  
    Parameters:
    None
  
    Returns:
    The type of control method for the drivers connection(i.e. Network, Serial,
    IR, or (none))
===============================================================================]]
function SetControlMethod()
	if (gCon ~= nil ) then 
	   gCon:destruct()
	   gCon = nil
	end

	if( gIsNetworkServerConnected) then
		-- connect to NetworkServer communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "NetworkServer") then
			gCon = NetworkServerConnectionBase:new()
			gCon:Initialize()
		end
		gCon:SetConnection(true, "NetworkServer")
	elseif (gIsNetworkConnected) then
		-- connect to Network communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Network") then
			gCon = NetworkConnectionBase:new(NETWORK_BINDING_ID, NETWORK_PORT)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "Network")
	elseif (gIsUrlConnected) then
		-- connect to URL communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "URL") then
			gCon = UrlConnectionBase:new()
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "URL")
	elseif (gIsSerialConnected) then
		-- connect to Serial communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Serial") then
			gCon = SerialConnectionBase:new(SERIAL_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
			gCon:InitPolling(tonumber(gPollingTimerInterval), "MINUTES", gCon)
		end
		gCon:SetConnection(true, "Serial")
	elseif (gIsIRConnected) then
		-- connect to IR communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "IR") then
			gCon = IRConnectionBase:new(IR_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "IR")
	else
		if (gCon ~= nil) then
			gCon:SetConnection(false, "(none)")
		end
		-- gCon = nil
	end

	gCon._CommandQueue:clear()
	gCon._Priority1CommandQueue:clear()
	gCon._Priority2CommandQueue:clear()
end

--[[=============================================================================
    ValidateControlMethod(controlMethod)
  
    Description:
    Identifies whether the specified control method has a valid connection
  
    Parameters:
    controlMethod(string) - The communication we are validating against
                            Valid types are (Network, Serial, and IR)
  
    Returns:
    true if the controlMethod specified has been connected, false otherwise.
===============================================================================]]
function ValidateControlMethod(controlMethod)
	local isValid = false

	if (controlMethod == "Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (controlMethod == "URL") and (gIsUrlConnected) then
		isValid = true
	elseif (controlMethod == "Serial") and (gIsSerialConnected) then
		isValid = true
	elseif (controlMethod == "IR") and (gIsIRConnected) then
		isValid = true
	end

	return isValid
end
 end)
package.preload['common.c4_network_connection'] = (function (...)
--[[=============================================================================
    Base for a network connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_device_connection_base"
require "lib.c4_log"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_network_connection = "2016.01.08"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30
DEFAULT_RECONNECT_INTERVAL_SECONDS = 5

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS
gNetworkReconnectInterval = DEFAULT_RECONNECT_INTERVAL_SECONDS

NetworkConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkConnectionBase:construct(BindingID, Port)
	self.superClass():construct()

	self._BindingID = BindingID
	self._Port = Port
	self._LastCheckin = 0
	self._IsOnline = false
	self._KeepAliveTimer = nil
end

function NetworkConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetConBase:Initialize")
	gControlMethod = "Network"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	self._KeepAliveTimer = c4_timer:new("PollingTimer", gNetworkKeepAliveInterval, "SECONDS", NetworkConnectionBase.OnKeepAliveTimerExpired, false, self)
	self._NetworkReconnectTimer = c4_timer:new("NetworkReconnectTimer", gNetworkReconnectInterval, "SECONDS", NetworkConnectionBase.OnNetworkReconnectTimerExpired, false, self)
end

function NetworkConnectionBase:ControlMethod()
	return "Network"
end

function NetworkConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)

			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end

function NetworkConnectionBase:ReceivedFromNetwork(idBinding, nPort, sData)
	self._LastCheckin = 0
	self:ReceivedFromCom(sData)
end

function NetworkConnectionBase:CheckNetworkConnectionStatus()
	if (self._IsConnected and (not self._IsOnline)) then
		LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(self._BindingID, self._Port)
		--C4:NetConnect(self._BindingID, self._Port)
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	end
end

function NetworkConnectionBase.OnKeepAliveTimerExpired(Instance)
	Instance._LastCheckin = Instance._LastCheckin + 1

	if(Instance._LastCheckin == 3) then
		LogWarn("Failed to receive poll responses... initiating network recovery mode...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		return
	elseif(Instance._LastCheckin > 4) then	
		Instance._LastCheckin = 4
	end

	if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function" and Instance._IsOnline) then
		SendKeepAlivePollingCommand()
	end

	Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
end

function NetworkConnectionBase.OnNetworkReconnectTimerExpired(Instance)
	if (Instance._IsConnected) then
		LogWarn("OnNetworkReconnectTimerExpired: Attempting to reactivate network connection...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		C4:NetConnect(Instance._BindingID, Instance._Port)

		if (gProjectorProxy._PowerState == "ON") then
		  Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	     else
		  LogWarn("The Projector is Powered Off, not restarting NetworkReconnectTimer.")
	     end
	else
		LogWarn("Cannot attempt to reactivate, the network connection is not bound")
		LogWarn("gProjectorProxy._PowerState = ".. gProjectorProxy._PowerState)
	end
end

function NetworkConnectionBase:SetOnlineStatus(IsOnline)
	self._IsOnline = IsOnline

	if(IsOnline) then
		self._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
		self._NetworkReconnectTimer:KillTimer()
		self._LastCheckin = 0
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "true")
		end

		self:SendNextCommand()
	else
		self._KeepAliveTimer:KillTimer()
		if (gProjectorProxy._PowerState == "ON") then
		  self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	     else
		  LogWarn("The Projector is Powered Off, not restarting NetworkReconnectTimer.")
	     end
	     if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
		  UpdateProperty("Connected To Network", "false")
	     end
	end
end

function ON_DRIVER_LATEINIT.c4_network_connection()
	-- Ensure existing connection is taken into consideration (useful on Driver Update)
	if (gIsNetworkConnected) then
		if (gCon ~= nil and gCon._BindingID ~= nil) then
			local tmp = C4:GetBindingAddress(gCon._BindingID)
			if (tmp ~= nil and string.len(tmp) > 0) then 
				OnNetworkBindingChanged(gCon._BindingID, true) 
			end
		end
	end
end

function NetworkConnectionBase:destruct()
    self.superClass():destruct()
    
    self._BindingID = nil
    self._Port = nil
    self._LastCheckin = nil
  
    if (self._KeepAliveTimer ~= nil) then 
	   self._KeepAliveTimer:destruct()
    end
    
    if (self._NetworkReconnectTimer ~= nil) then 
	   self._NetworkReconnectTimer:destruct()
    end

    if (self._PollingTimer ~= nil) then 
	   self._PollingTimer:destruct()
    end
    
    gTicketIdMap = {}
    
end end)
package.preload['common.c4_notify'] = (function (...)
--[[=============================================================================
    Notification Functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_notify = "2016.01.08"
end

--[[=============================================================================
    SendNotify(notifyText, tParams, bindingID)

    Description
    Forwards a notification to the proxy with a list of parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    tParams(table)     - Table of key value pairs that hold the the parameters
                         and their values used in the proxy function
    bindingID(int)     - The requests binding id

    Returns
    Nothing
===============================================================================]]
function SendNotify(notifyText, tParams, bindingID)
	C4:SendToProxy(bindingID, notifyText, tParams, "NOTIFY")
end

--[[=============================================================================
    SendSimpleNotify(notifyText, ...)

    Description
    Forwards a notification to the proxy with no parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    bindingID(int)     - Optional parameter containing the requests binding id,
                         if not specified then the DEFAULT_PROXY_ID is given.

    Returns
    Nothing
===============================================================================]]
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end end)
package.preload['common.c4_command'] = (function (...)
--[[=============================================================================
    Functions for handling and executing commands and actions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_command = "2016.01.08"
end

--[[=============================================================================
    ExecuteCommand(sCommand, tParams)

    Description
    Function called by Director when a command is received for this DriverWorks
    driver. This includes commands created in Composer programming.

    Parameters
    sCommand(string) - Command to be sent
    tParams(table)   - Lua table of parameters for the sent command

    Returns
    Nothing
===============================================================================]]
function ExecuteCommand(sCommand, tParams)
	LogTrace("ExecuteCommand(" .. sCommand .. ")")
	LogInfo(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")
	local status, ret

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		status, ret = pcall(EX_CMD[sCommand], tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		status, ret = pcall(EX_CMD[trimmedCommand], tParams)
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
		status = true
	else
		LogInfo("ExecuteCommand: Unhandled command = " .. sCommand)
		status = true
	end
	
	if (not status) then
		LogError("LUA_ERROR: " .. ret)
	end
	
	return ret -- Return whatever the function returns because it might be xml, a return code, and so on
end

--[[=============================================================================
    EX_CMD.LUA_ACTION(tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    tParams(table) - Lua table of parameters for the command option

    Returns
    Nothing
===============================================================================]]
function EX_CMD.LUA_ACTION(tParams)
	if (tParams ~= nil) then
		for cmd, cmdv in pairs(tParams) do
			if (cmd == "ACTION" and cmdv ~= nil) then
				local status, err = pcall(LUA_ACTION[cmdv], tParams)
				if (not status) then
					LogError("LUA_ERROR: " .. err)
				end
				break
			end
		end
	end
end

--[[=============================================================================
    ReceivedFromProxy(idBinding, sCommand, tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    idBinding(int)   - Binding ID of the proxy that sent a BindMessage to the
                       DriverWorks driver.
    sCommand(string) - Command that was sent
    tParams(table)   - Lua table of received command parameters

    Returns
    Nothing
===============================================================================]]
function ReceivedFromProxy(idBinding, sCommand, tParams)

	if (sCommand ~= nil) then

		-- initial table variable if nil
		if (tParams == nil) then
			tParams = {}
		end
		
		LogTrace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function PRX_CMD." .. sCommand .. "()")
		LogInfo(tParams)

		if ((PRX_CMD[sCommand]) ~= nil) then
			local status, err = pcall(PRX_CMD[sCommand], idBinding, tParams)
			if (not status) then
				LogError("LUA_ERROR: " .. err)
			end
		else
			LogInfo("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

--[[
	This function is called when a UI (Navigator) requests data, and
	calls the function requested.
--]]
function UIRequest(sRequest, tParams)
	local ret = ""

	if (sRequest ~= nil) then
		tParams = tParams or {}   -- initial table variable if nil
		LogTrace("UIRequest(): " .. sRequest .. "; Call Function UI_REQ." .. sRequest .. "()")
		LogInfo(tParams)

		if (UI_REQ[sRequest]) ~= nil then
			ret = UI_REQ[sRequest](tParams)
		else
			LogWarn("UIRequest: Unhandled request = " .. sRequest)
		end
	end

	return ret
end
 end)
package.preload['common.c4_property'] = (function (...)
--[[=============================================================================
    Function for changing properties

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_property = "2016.01.08"
end

--[[=============================================================================
    OnPropertyChanged(sProperty)

    Description
    Function called by Director when a property changes value. The value of the
    property that has changed can be found with: Properties[sName]. Note that
    OnPropertyChanged is not called when the Property has been changed by the
    driver calling the UpdateProperty command, only when the Property is changed
    by the user from the Properties Page. This function is called by Director
    when a property changes value.

    Parameters
    sProperty(string) - Name of property that has changed.

    Returns
    Nothing
===============================================================================]]
function OnPropertyChanged(sProperty)
	local propertyValue = Properties[sProperty]

	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	end

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	local status = true
	local err = ""

	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[sProperty], propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[trimmedProperty], propertyValue)
	end

	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)
	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end
 end)
package.preload['common.c4_init'] = (function (...)
--[[=============================================================================
    Initial driver initialization and destruction functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "common.c4_property"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_init = "2016.01.08"
end

--[[=============================================================================
    OnDriverInit()

    Description
    Invoked by director when a driver is loaded. This API is provided for the
    driver developer to contain all of the driver objects that will require
    initialization.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverInit()
	gInitializingDriver = true
	C4:ErrorLog("INIT_CODE: OnDriverInit()")

	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_EARLY_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property
	-- global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		C4:ErrorLog("INIT_CODE: Calling OnPropertyChanged - " .. k .. ": " .. v)
		local status, err = pcall(OnPropertyChanged, k)
		if (not status) then
			C4:ErrorLog("LUA_ERROR: " .. err)
		end
	end

	gInitializingDriver = false
end

--[[=============================================================================
    OnDriverLateInit()

    Description
    Invoked by director after all drivers in the project have been loaded. This
    API is provided for the driver developer to contain all of the driver
    objects that will require initialization after all drivers in the project
    have been loaded.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[=============================================================================
    OnDriverDestroyed()
    Function called by Director when a driver is removed. Release things this
    driver has allocated such as timers.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	
	-- Call all ON_DRIVER_DESTROYED functions
	for k, v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end end)
package.preload['lib.c4_timer'] = (function (...)
--[[=============================================================================
    c4_timer Class

    Copyright 2017 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_timer = "2017.05.03"
end

c4_timer = inheritsFrom(nil)

function c4_timer:construct(name, interval, units, Callback, repeating, CallbackParam)
	self._name = name
	self._timerID = TimerLibGetNextTimerID()
	self._interval = interval
	self._units = units
	self._repeating = repeating or false
	self._Callback = Callback
	self._CallbackParam = CallbackParam or ""
	self._id = 0

	gTimerLibTimers[self._timerID] = self
	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("Created timer " .. self._name)
	end
end

function c4_timer:destruct()
    self:KillTimer()
    gTimerLibTimers[self._timerID] = nil
end


function c4_timer:StartTimer(...)
	c4_timer:KillTimer()

	-- optional parameters (interval, units, repeating)
	if ... then
		local interval = select(1, ...)
		local units = select(2, ...)
		local repeating = select(3, ...)

		self._interval = interval or self._interval
		self._units = units or self._units
		self._repeating = repeating or self._repeating
	end

	if (tonumber(self._interval) > 0) then
		if (LOG ~= nil and type(LOG) == "table") then
			LogTrace("Starting Timer: " .. self._name)
		end

		self._id = C4:AddTimer(self._interval, self._units, self._repeating)
	end
end

function c4_timer:KillTimer()
	if (self._id) then
		self._id = C4:KillTimer(self._id)
	end
end

function c4_timer:TimerStarted()
	return (self._id ~= 0)
end

function c4_timer:TimerStopped()
	return (self._id == 0)
end

function c4_timer:GetTimerInterval()
	return (self._interval)
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.c4_timer()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.c4_timer()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[=============================================================================
    OnTimerExpired(idTimer)

    Description:
    Function called by Director when the specified Control4 timer expires.

    Parameters:
    idTimer(string) - Timer ID of expired timer.
===============================================================================]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._CallbackParam)
			end
		end
	end
end

--[[=============================================================================
    CreateTimer(name, interval, units, callback, repeating, callbackParam)

    Description: 
    Creates a named timer with the given attributes

    Parameters:
    name(string)       - The name of the timer being created
    interval(int)      - The amount of the given time between calls to the
                         timers callback function
    units(string)      - The time of time interval used (e.g. MILLSECONDS, SECONDS, MINUTES, HOURS)
    callback(string)   - The function to call when the timer expires
    repeating(bool)    - Parameter indicating whether the timer should be
                         called repeatedly until cancelled
    callbackParam(...) - Parameters to be passed to the callback function

    Returns:
    A handle to the timer
===============================================================================]]
function CreateTimer(name, interval, units, callback, repeating, callbackParam)
	timer = c4_timer:new(name, interval, units, callback, repeating, callbackParam)
	return timer
end

--[[=============================================================================
    StartTimer(handle, ...)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer)   - Handle to a created timer object
    interval(int)   - The amount of the given time between calls to the
                      timers callback function
    units(string)   - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    repeating(bool) - Parameter indicating whether the timer should be
                      called repeatedly until cancelled

    Returns:
    None
===============================================================================]]
function StartTimer(handle, ...)
	handle:StartTimer(...)
end

--[[=============================================================================
    KillTimer(handle)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    None
===============================================================================]]
function KillTimer(handle)
	handle:KillTimer()
end

--[[=============================================================================
    TimerStarted(handle)

    Description: 
    Identifies whether a timer has been started or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been started, or false otherwise
===============================================================================]]
function TimerStarted(handle)
	return handle:TimerStarted()
end

--[[=============================================================================
    TimerStopped(handle)

    Description: 
    Identifies whether a timer has been stopped or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been stopped, or false otherwise
===============================================================================]]
function TimerStopped(handle)
	return handle:TimerStopped()
end

--[[=============================================================================
    GetTimerInterval(handle)

    Description: 
    Gets the interval setting of the given timer

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns the interval setting of the given timer
===============================================================================]]
function GetTimerInterval(handle)
	return handle:GetTimerInterval()
end

--[[=============================================================================
    c4_timer Unit Tests
===============================================================================]]
function __test_c4_timer()
	require "test.C4Virtual"
	require "lib.c4_log"
	require "common.c4_init"

	OnDriverInit()

	local LOG = c4_log:new("test_c4_timer")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	function OnTestTimerExpired()
		c4Timer:KillTimer()
	end

	-- create an instance of the timer
	c4Timer = c4_timer:new("Test", 45, "MINUTES", OnTestTimerExpired)

	assert(c4Timer._id == 0, "_id is not equal to '0' it is: " .. c4Timer._id)
	c4Timer:StartTimer()
	assert(c4Timer._id == 10001, "_id is not equal to '10001' it is: " .. c4Timer._id)
	assert(c4Timer:TimerStarted() == true, "TimerStarted is not equal to true it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == false, "TimerStopped is not equal to false it is: " .. tostring(c4Timer:TimerStopped()))
	OnTimerExpired(c4Timer._id)
	assert(c4Timer:TimerStarted() == false, "TimerStarted is not equal to false it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == true, "TimerStopped is not equal to true it is: " .. tostring(c4Timer:TimerStopped()))
end end)
package.preload['lib.c4_object'] = (function (...)
--[[=============================================================================
    c4_object Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_object = "2016.01.08"
end

function inheritsFrom( baseClass )
	local new_class = {}
	local class_mt = { __index = new_class }

	function new_class:create(...)
		local newinst = {}

		setmetatable( newinst, class_mt )

		-- Call the constructor when we create this class
		if newinst.construct then
			-- Allow returning a different obj than self. This allows for readonly tables etc...
			newinst = newinst:construct(...) or newinst
		end

		return newinst
	end

	if nil ~= baseClass then
		setmetatable( new_class, { __index = baseClass } )
	end

	--[[=============================================================================
	    Implementation of additional OO properties starts here 
	===============================================================================]]
	
	-- Return the class object of the instance
	function new_class:class()
		return new_class
	end

	--[[=============================================================================
	    Return the super class object of the instance. 

	    Note Calling methods on the base class itself will modify
	    the base table's static properties. In order to have call
	    the base class' methods and have them modify the current object
	    use super() or superAsSelf(). 
	===============================================================================]]
	function new_class:superClass()
		return baseClass
	end

	--[[=============================================================================
	    Returns a table that allows calling of the base class's method
	    while maintaining the objects state as the modified state of the base
	    class' methods. For example consider the following statements (order matters):

	    -- The child sees the parents property if the child hasn't overriden the property
	       obj:superClass().id = "parent"
	       obj.id == "parent" -- true

	    -- Setting the property on the child overrides (hides) the parents property
	       obj.id = "child"
	       obj.id == "child" -- true

	    -- The super() method pass
	       obj:super().id == "parent" -- true
	       obj:super().id = "child"
	       obj:super().id == "parent" -- still true
	       obj.id == "child" -- still true
	===============================================================================]]
	function new_class:super()
		local holder = {}

		holder.child = self
		holder.parent = baseClass

		local mt = {}
		mt.__index = function(table, index)
			if table.parent[index] then
				return table.parent[index]
			else 
				return table.child[index]
			end
		end

		-- Only set the new values to the child.
		mt.__newindex = function(table, key, value)
			table.child[key] = value
		end

		mt.__tostring = function(table)
			return tostring(table.child)
		end

		setmetatable(holder, mt)
		return holder
	end

	new_class.new = new_class.create

	--[[=============================================================================
	    Return true if the caller is an instance of theClass
	===============================================================================]]
	function new_class:isa( theClass )
		local b_isa = false
		local cur_class = new_class

		while ( nil ~= cur_class ) and ( false == b_isa ) do
			if cur_class == theClass then
				b_isa = true
			else
				cur_class = cur_class:superClass()
			end
		end

		return b_isa
	end

	return new_class
end

--[[=============================================================================
    Inheritance unit tests
===============================================================================]]
function __test_inheritance()
	local b = inheritsFrom(nil)

	b.construct = function(self, msg)
		self._msg = msg
	end

	local t = inheritsFrom(b)
	t.construct = function(self, msg)
		self:super():construct(msg)
	end

	t1 = t:new("t1")
	t2 = t:new("t2")
	assert(t1._msg == "t1", "t1 message is not equal to 't1' it''s: " .. t1._msg)
	assert(t2._msg == "t2", "t2 message is not equal to 't2' it''s: " .. t2._msg)    
	assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");
	assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

	t1:superClass().id = "parent"
	assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

	-- Setting the property on the child overrides (hides) the parents property
	t1.id = "child"
	assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

	-- The super() method maintains the self pointer to the child and not to the base
	assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
	t1:super().id = "child1"
	assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
	assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")
end end)
package.preload['lib.c4_log'] = (function (...)
--[[=============================================================================
    c4_log Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_object"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_log = "2016.01.08"
end

c4_log = inheritsFrom(nil)

function c4_log:construct(logName)
	self._logLevel = tonumber(string.sub(Properties['Log Level'] or "", 1, 1)) or 5
	self._outputPrint = Properties['Log Mode']:find("Print") ~= nil
	self._outputC4Log = Properties['Log Mode']:find("Log") ~= nil
	self._logName = logName or ""

	-- make sure Property is up to date (no harm if absent)
	C4:UpdateProperty("Log Level", Properties['Log Level'])
end

function c4_log:SetLogLevel(level)
	self._logLevel = tonumber(string.sub(level or "", 1, 1)) or self._logLevel
end

function c4_log:LogLevel()
	return self._logLevel
end

function c4_log:OutputPrint(value)
	self._outputPrint = value
end

function c4_log:OutputC4Log(value)
	self._outputC4Log = value
end

function c4_log:SetLogName(logName)

	if (logName == nil or logName == "") then
		logName = ""
	else
		logName = logName .. ": "
	end

	self._logName = logName
end

function c4_log:LogName()
	return self._logName
end

function c4_log:Enabled()
	return (self._outputPrint or self._outputC4Log)
end

function c4_log:PrintEnabled()
	return self._outputPrint
end

function c4_log:C4LogEnabled()
	return self._outputC4Log
end

function c4_log:CreateTableText(tValue, tableText)
	tableText = tableText or ""

	if (type(tValue) == "table") then

		tableText = tableText .. "{"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ","
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableText(v, tableText)
				tableText = tableText .. ","
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ","
			end
		end

		tableText = tableText .. "}"
	end

	return tableText
end

function InsertIndent(indentLevel)
	local indentStr = ""

	for i=1, indentLevel do
		indentStr = indentStr .. "\t"
	end

	return indentStr
end

function c4_log:CreateTableTextFormatted(tValue, tableText, indentLevel)
	tableText = tableText or ""
	indentLevel = indentLevel or 0

	if (type(tValue) == "table") then

		indentLevel = indentLevel + 1
		tableText = tableText .. "{\n"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. InsertIndent(indentLevel) .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. InsertIndent(indentLevel) .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ",\n"
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',\n"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableTextFormatted(v, tableText, indentLevel)
				tableText = tableText .. ",\n"
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ",\n"
			end
		end

		indentLevel = indentLevel - 1
		tableText = tableText .. InsertIndent(indentLevel) .. "}"
	end

	return tableText
end

MAX_TABLE_LEVELS = 10
function c4_log:PrintTable(tValue, tableText, sIndent, level)
	tableText = tableText or ""
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				if (tableText == "") then
					tableText = sIndent .. tostring(k) .. ":  " .. tostring(v)
					if (sIndent == ".   ") then sIndent = "    " end
				else
					tableText = tableText .. "\n" .. sIndent .. tostring(k) .. ":  " .. tostring(v)
				end
				if (type(v) == "table") then
					tableText = self:PrintTable(v, tableText, sIndent .. "   ", level)
				end
			end
		else
			tableText = tableText .. "\n" .. sIndent .. tostring(tValue)
		end
	end
	
	return tableText
end

function c4_log:LogTable(tValue, sIndent, level)
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				C4:ErrorLog(self._logName .. sIndent .. tostring(k) .. ":  " .. tostring(v))
				if (type(v) == "table") then
					self:LogTable(v, sIndent .. "   ", level)
				end
			end
		else
			C4:ErrorLog(self._logName .. sIndent .. tValue)
		end
	end
end

function c4_log:Print(logLevel, sLogText)

	if (self._logLevel >= logLevel) then
		if (type(sLogText) == "table") then
			if (self._outputPrint) then
				print (self:PrintTable(sLogText, tableText, ".   ", 0))
			end

			if (self._outputC4Log) then
				self:LogTable(sLogText, "   ", 0)
			end

			return
		end

		if (self._outputPrint) then
			print (sLogText)
		end

		if (self._outputC4Log) then
			C4:ErrorLog(self._logName .. tostring(sLogText))
		end
	end
end

function c4_log:Fatal(sLogText, ...)
	self:LogOutput(0, sLogText, ...)
end

function c4_log:Error(sLogText, ...)
	self:LogOutput(1, sLogText, ...)
end

function c4_log:Warn(sLogText, ...)
	self:LogOutput(2, sLogText, ...)
end

function c4_log:Info(sLogText, ...)
	self:LogOutput(3, sLogText, ...)
end

function c4_log:Debug(sLogText, ...)
	self:LogOutput(4, sLogText, ...)
end

function c4_log:Trace(sLogText, ...)
	self:LogOutput(5, sLogText, ...)
end

function c4_log:LogOutput(level, sLogText, ...)
	if (LogEnabled()) then
		if (type(sLogText) == "string") then
			sLogText = string.format(sLogText, ...)
		end

		self:Print(level, sLogText)
	end
end

--[[=============================================================================
    c4_log wrapper functions
===============================================================================]]
function TryLog(level, sLogText, ...)
	LOG:LogOutput(level, sLogText, ...)
end

--[[=============================================================================
    SetLogLevel(level)

    Description: 
    Sets the desired log level to view

    Parameters:
    level(int) - The logging level to set the message to
                 0 = Fatal
                 1 = Error
                 2 = Warn
                 3 = Info
                 4 = Debug
                 5 = Trace

    Returns:
    None
===============================================================================]]
function SetLogLevel(level)
	LOG:SetLogLevel(level)
end

--[[=============================================================================
    LogLevel()

    Description: 
    Returns the currently set log level

    Parameters:
    None

    Returns:
    The current log level
        0 = Fatal
        1 = Error
        2 = Warn
        3 = Info
        4 = Debug
        5 = Trace
===============================================================================]]
function LogLevel()
	return LOG:LogLevel()
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputPrint(value)
	LOG:OutputPrint(value)
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages to file or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputC4Log(value)
	LOG:OutputC4Log(value)
end

--[[=============================================================================
    SetLogName(logName)

    Description: 
    Sets the name of the log file where the messages will be written to

    Parameters:
    logName(string) - Sets the name of the log to write messages to

    Returns:
    None
===============================================================================]]
function SetLogName(logName)
	LOG:SetLogName(logName)
end

--[[=============================================================================
    LogName(logName)

    Description: 
    Gets the name of the log file where the messages will be written to

    Parameters:
    None

    Returns:
    The value of the log file that has been set
===============================================================================]]
function LogName()
	return LOG:LogName()
end

--[[=============================================================================
    LogEnabled()

    Description: 
    Identifies if logging or print has been enabled

    Parameters:
    None

    Returns:
    true if either logging or print has been enabled, false otherwise
===============================================================================]]
function LogEnabled()
	return LOG:Enabled()
end

--[[=============================================================================
    PrintEnabled()

    Description: 
    Gets the state of print output

    Parameters:
    None

    Returns:
    true if print has been enabled, false otherwise
===============================================================================]]
function PrintEnabled()
	return LOG:PrintEnabled()
end

--[[=============================================================================
    C4LogEnabled()

    Description: 
    Gets the state of logging

    Parameters:
    None

    Returns:
    true if logging has been enabled, false otherwise
===============================================================================]]
function C4LogEnabled()
	return LOG:C4LogEnabled()
end

--[[=============================================================================
    LogFatal(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Fatal(0) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogFatal(sLogText, ...)
	local status, err = pcall(TryLog, 0, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogFatal failed: " .. err)
	end
end

--[[=============================================================================
    LogError(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Error(1) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogError(sLogText, ...)
	local status, err = pcall(TryLog, 1, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogError failed: " .. err)
	end
end

--[[=============================================================================
    LogWarn(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Warn(2) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogWarn(sLogText, ...)
	local status, err = pcall(TryLog, 2, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogWarn failed: " .. err)
	end
end

--[[=============================================================================
    LogInfo(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Info(3) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogInfo(sLogText, ...)
	local status, err = pcall(TryLog, 3, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogInfo failed: " .. err)
	end
end

--[[=============================================================================
    LogDebug(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Debug(4) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogDebug(sLogText, ...)
	local status, err = pcall(TryLog, 4, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogDebug failed: " .. err)
	end
end

--[[=============================================================================
    LogTrace(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Trace(5) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogTrace(sLogText, ...)
	local status, err = pcall(TryLog, 5, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogTrace failed: " .. err)
	end
end

function dbgPrint(buf)
	if (LOG:PrintEnabled()) then
		print (buf)
	end
end

function dbgHexdump(buf)
	hexdump(buf, dbgPrint)
end

--[[=============================================================================
    c4_log unit tests
===============================================================================]]
function __test_c4_log()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	assert(LOG:LogName() == "test_c4_log", "_logName is not equal to 'test_c4_log' it is: " .. LOG:LogName())

	-- Test setting log level
	LOG:SetLogLevel("2 - Warning")
	assert(LOG:LogLevel() == 2, "_logLevel is not equal to '2' it is: " .. LOG:LogLevel())

	LOG:SetLogLevel(3)
	assert(LOG:LogLevel() == 3, "_logLevel is not equal to '3' it is: " .. LOG:LogLevel())

	-- Test enabling logs
	LOG:OutputPrint(false)
	assert(LOG:PrintEnabled() == false, "_outputPrint is not equal to 'false' it is: " .. tostring(LOG:PrintEnabled()))

	LOG:OutputC4Log(true)
	assert(LOG:C4LogEnabled() == true, "_outputC4Log is not equal to 'true' it is: " .. tostring(LOG:C4LogEnabled()))

	LOG:SetLogLevel(4)
	LogTrace("***** This is a test *****")
end

function __test_CreatTableText()
	local tTest = {}

	tTest[1] = {}
	tTest[2] = {}
	tTest[3] = 30
	tTest[4] = "Forty"
	
	LogTrace("----- tText -----")
	LogTrace(tTest)

	local tTest2 = { One = {},
					 Two = {},
					 Three = 30,
					 Four = "Forty" }
	LogTrace("----- tText2 -----")
	LogTrace(tTest2)
	
	local tTest3 = { [1] = {},
					 [2] = {},
					 [3] = 30,
					 [4] = "Forty" }
	LogTrace("----- tText3 -----")
	LogTrace(tTest3)

	local tTest4 = { [1] = {},
					 Two = {},
					 [3] = 30,
					 [4] = "Forty",
					 Five = "Fifty" }
	LogTrace("----- tText4 -----")
	LogTrace(tTest4)

	local tableText = LOG:CreateTableText(tTest4)
	LogTrace("----- tableText -----")
	LogTrace(tableText)
	
	--local tNew = {[1] = {},[3] = 30,[4] = 'Forty',Five = 'Fifty',Two = {},}
	--LogTrace(tNew)
end

function __TestCreateTableTextFormatted()
	require "test.C4Virtual"
	
	local LOG = c4_log:new("test_c4_log")
	local tButtons = {
		Name = 'heat',
		Attributes = {},
		ChildNodes = {
			[1] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '51',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Pool Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'POOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[2] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '53',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Spa Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'SPAHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[3] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '54',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {Value = 'Pool Solar Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {}
					},
					[3] = {
						Value = 'SOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				}
			}
		}
	}

	print(LOG:CreateTableTextFormatted(tButtons))
end end)
package.preload['common.c4_common'] = (function (...)
--[[=============================================================================
    ON_INIT, Timer,s and Property management functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]
require "common.c4_driver_declarations"
require "lib.c4_log"
require "lib.c4_timer"

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_common = "2016.01.08"
end

--[[=============================================================================
    Create and Initialize Logging
===============================================================================]]
function ON_DRIVER_EARLY_INIT.c4_common()
	-- Create a logger
	LOG = c4_log:new("Template_c4z Change Name")
end

function ON_DRIVER_INIT.c4_common()
	-- Create Log Timer
	gC4LogTimer = c4_timer:new("Log Timer", 45, "MINUTES", OnLogTimerExpired)
end

--[[=============================================================================
    Log timer callback function
===============================================================================]]
function OnLogTimerExpired()
	LogWarn("Turning Log Mode Off (timer expired)")
	gC4LogTimer:KillTimer()
	
	C4:UpdateProperty("Log Mode", "Off")
	OnPropertyChanged("Log Mode")
end

gForceLogging = false
function ON_PROPERTY_CHANGED.LogMode(propertyValue)
	gC4LogTimer:KillTimer()
	
	if (gForceLogging) then
		LOG:OutputPrint(true)
		LOG:OutputC4Log(true)
	else
		LOG:OutputPrint(propertyValue:find("Print") ~= nil)
		LOG:OutputC4Log(propertyValue:find("Log") ~= nil)
		if (propertyValue == "Off") then
			return
		end
		
		gC4LogTimer:StartTimer()
	end
end

function ON_PROPERTY_CHANGED.LogLevel(propertyValue)
	if (gForceLogging) then
		LOG:SetLogLevel("5 - Debug")
	else
		LOG:SetLogLevel(propertyValue)
	end
end

--[[=============================================================================
    Print Template Versions
===============================================================================]]
function TemplateVersion()
	print ("\nTemplate Versions")
	print ("-----------------------")
	for k, v in pairs(TEMPLATE_VERSION) do
		print (k .. " = " .. v)
	end
	
	print ("")
end end)
package.preload['common.c4_driver_declarations'] = (function (...)
--[[=============================================================================
    Driver Declarations used to call startup routines, teardown routines, and 
    other basic functions of the drivers operation

    Copyright 2017 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Template Version Table
TEMPLATE_VERSION = {}
TEMPLATE_VERSION.c4_driver_declarations = "2017.04.25"

-- Command Handler Tables
EX_CMD = {}
PRX_CMD = {}
UI_REQ = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}
PROG_CONDITIONAL = {}


--[[=============================================================================
    Tables of functions
    The following tables are function containers that are called within the
    following functions:

    OnDriverInit()
        First calls all functions contained within ON_DRIVER_EARLY_INIT table
        then calls all functions contained within ON_DRIVER_INIT table

    OnDriverLateInit()
        Calls all functions contained within ON_DRIVER_LATEINIT table

    OnDriverDestroyed()
        Calls all functions contained within ON_DRIVER_DESTROYED table

    OnPropertyChanged()
        Calls all functions contained within ON_PROPERTY_CHANGED table
===============================================================================]]
ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001 end)
--[[=============================================================================
    Basic Template for PROJECTOR Driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]------------
require "common.c4_driver_declarations"
require "common.c4_common"
require "common.c4_init"
require "common.c4_property"
require "common.c4_command"
require "common.c4_notify"
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_utils"
require "lib.c4_timer"
require "actions"
require "device_specific_commands"
require "device_messages"
require "projector_init"
require "properties"
require "proxy_commands"
require "connections"
require "projector.projector_proxy_class"
require "projector.projector_proxy_commands"
require "projector.projector_proxy_notifies"


-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.driver = "2015.03.31"
end

--[[=============================================================================
    Constants
===============================================================================]]
PROJECTOR_PROXY_BINDINGID = 5001

--[[
    Device Volume Range
    TODO: edit "MIN_DEVICE_LEVEL" & "MAX_DEVICE_LEVEL" values based upon the protocol specification for volume range.
    If zones have volume ranges that vary, convert these constants into a table indexed by the mod value of the output connection 
    and then update all references to handle the table structure.
--]]
MIN_DEVICE_LEVEL = 0
MAX_DEVICE_LEVEL = 100	


--[[=============================================================================
    Initialization Code
===============================================================================]]
function ON_DRIVER_EARLY_INIT.main()
    --load JSON library
    require "json"
    JSON=(loadstring(json.JSON_LIBRARY_CHUNK))()	
end

function ON_DRIVER_INIT.main()

end

function ON_DRIVER_LATEINIT.main()
	C4:urlSetTimeout(20)
	DRIVER_NAME = C4:GetDriverConfigInfo("name")
	
	SetLogName(DRIVER_NAME)
end

function ON_DRIVER_EARLY_INIT.projector_driver()

end

function ON_DRIVER_INIT.projector_driver()
    -- TODO: Modify tVolumeRamping to have on entry per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {["state"] = false, ["mode"] = ""}
	
    -- Create an instance of the PROJECTOR Proxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local  bProcessesDeviceMessages = true
    local bUsePulseCommandsForVolumeRamping = false
    if (Properties["Ramping Method"] == "Pulse Commands") then bUsePulseCommandsForVolumeRamping = true end
    gProjectorProxy = ProjectorProxy:new(PROJECTOR_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping, bUsePulseCommandsForVolumeRamping)
    gProjectorProxy:Initialize()
    
    gHandleACKDelayTimer = c4_timer:new("HandleACKDelayTimer", 500, "MILLISECONDS", OnHandleACKDelayTimerExpired)
    gNetworkInitTimer = c4_timer:new("NetworkInitTimer", 5, "SECONDS", OnNetworkInitTimerExpired)
    
    gPowerOnTimer = c4_timer:new("PowerOnTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnTimerExpired)
    gPowerOffTimer = c4_timer:new("PowerOffTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffTimerExpired)     
    gSendPowerOnAfterPowerOffDelayTimer = c4_timer:new("SendPowerOnAfterPowerOffDelayTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnSendPowerOnAfterPowerOffDelayTimerExpired)
    gQueryPowerDelayTimer = c4_timer:new("QueryPowerDelayTimer", 1, "SECONDS", OnQueryPowerDelayTimerExpired)
    
    gSetInputDelayTimer = c4_timer:new("SetInputDelayTimer", 2, "SECONDS", OnSetInputDelayTimerExpired, command)
    
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
	
	--[[
	For the "Volume Curve" method, tVolumeCurve is used to store volume level values that will be used to build volume commands during volume ramping. Specifically, they are used in GetNextVolumeCurveValue() which is called from the ContinueVolumeRamping() function.  Property values for "Volume Ramping Steps" and "Volume Ramping Slope" can be adjusted to get a smooth volume ramping from low to high volume.
	--]]
end

function getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    local steps = tonumber(Properties["Volume Ramp Steps"])
    local slope = tonumber(Properties["Volume Ramp Slope"])
    local tCurve = gProjectorProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    
    return tCurve 
end

function ON_DRIVER_LATEINIT.projector_driver()
    if (DRIVER_CONTROL == "ir" or DRIVER_CONTROL == "serial") then
	   C4:SetPropertyAttribs ("Network Keep Alive Interval Seconds", 1)
	   C4:SetPropertyAttribs ("MAC Address", 1)
	   C4:SetPropertyAttribs ("IP Address", 1)
	   C4:SetPropertyAttribs ("Connected To Network", 1)
    end 
    C4:SetPropertyAttribs ("== Volume Ramping Settings ==", 1)
    C4:SetPropertyAttribs ("Ramping Method", 1)
    C4:SetPropertyAttribs ("Volume Ramp Steps", 1)
    C4:SetPropertyAttribs ("Volume Ramp Slope", 1)
    C4:SetPropertyAttribs ("Volume Ramp Delay Milliseconds", 1)

end

--[[=============================================================================
    Driver Code
===============================================================================]]

gSTX = tohex('A9')
gETX = tohex('9A')

function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then 
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
		  command_delay = 4
		  delay_units = "SECONDS"
		
		  --local data = JSON:encode(tData)
		  stx = ""
		  etx = "\r\n"
		  cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		--stx = ""
		--etx = "\r"
		--cmd = stx .. command .. etx
		
		stx = gSTX
		etx = gETX

		cmd = stx .. command .. string.pack('b',GetChecksum(command)) .. etx
		
		
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    
    if (command == "") then
	   LogWarn("PackAndQueuePriority1Command(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
	     --using ACK, so the delay values are for the WaitResponseTimer (timeout) 
	     command_delay = 4
	     delay_units = "SECONDS"
		
	     stx = ""
	     etx = "\r\n"
	     cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		stx = gSTX
		etx = gETX

		cmd = stx .. command .. string.pack('b',GetChecksum(command)) .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    LogTrace("PackAndQueuePriority1Command(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end

function GetChecksum(s)
-- Ensure no nil command for checksum
  s=s or ""
  local cs = 0
  s:gsub('.', function (c) cs = bit.bor(cs,string.byte(c)) end)
  return cs
end